from copy import deepcopy
import pytest

from traw.models import CaseType

ID = 1234
NAME = 'mock case type name'


@pytest.fixture()
def case_type(client):
    content = {"id": ID,
               "is_default": True,
               "name": NAME}
    return CaseType(client, deepcopy(content))


def test_id(case_type):
    assert case_type.id == ID


def test_is_default(case_type):
    assert case_type.is_default is True


def test_name(case_type):
    assert case_type.name == NAME



from copy import deepcopy
from datetime import datetime as dt

import pytest

from traw.models import Project, Suite

COMPLETED_ON = 1469726522
DESCRIPTION = "mock suite description"
ID = 123
NAME = 'mock name'
PROJECT_ID = 15
URL = 'mock url'


@pytest.fixture()
def new_suite(client):
    return Suite(client)


@pytest.fixture()
def suite(client):
    content = {"completed_on": COMPLETED_ON,
               "description": DESCRIPTION,
               "id": ID,
               "is_baseline": False,
               "is_completed": False,
               "is_master": False,
               "name": NAME,
               "project_id": PROJECT_ID,
               "url": URL}
    return Suite(client, deepcopy(content))


def test_completed_on(new_suite, suite):
    assert new_suite.completed_on is None
    assert isinstance(suite.completed_on, dt)
    assert suite.completed_on == dt.fromtimestamp(COMPLETED_ON)


def test_description_get(new_suite, suite):
    assert new_suite.description is None
    assert suite.description == DESCRIPTION


def test_description_set(new_suite, suite):
    new_suite.description = "New description 1"
    assert new_suite.description == "New description 1"

    suite.description = "New description 2"
    assert suite.description == "New description 2"


def test_description_set_exc(new_suite):
    with pytest.raises(TypeError) as exc:
        new_suite.description = 1234

    assert str(str) in str(exc)
    assert str(int) in str(exc)


def test_is_baseline(new_suite, suite):
    assert new_suite.is_baseline is False
    assert suite.is_baseline is False


def test_is_completed(new_suite, suite):
    assert new_suite.is_completed is False
    assert suite.is_completed is False


def test_is_master(new_suite, suite):
    assert new_suite.is_master is False
    assert suite.is_master is False


def test_name_get(new_suite, suite):
    assert new_suite.name is None
    assert suite.name == NAME


def test_name_set(new_suite, suite):
    assert new_suite.name is None
    new_suite.name = "New Suite Name"
    assert new_suite.name == "New Suite Name"

    assert suite.name == NAME
    suite.name = "New Suite Name"
    assert suite.name == "New Suite Name"


def test_name_set_exc(new_suite):
    with pytest.raises(TypeError) as exc:
        new_suite.name = 1234

    assert str(str) in str(exc)
    assert str(int) in str(exc)


def test_project_get(new_suite, suite):
    assert new_suite.project is None
    assert isinstance(suite.project, Project)

    suite.client.api.project_by_id.return_value = {'id': PROJECT_ID}

    assert suite.project.id == PROJECT_ID
    suite.client.api.project_by_id.assert_called_with(PROJECT_ID)


def test_project_set(new_suite):
    project = Project(None, {'id': PROJECT_ID})
    new_suite.project = project

    assert new_suite._content['project_id'] == PROJECT_ID


def test_project_set_exc(new_suite):
    with pytest.raises(TypeError) as exc:
        new_suite.project = 1234

    assert str(Project) in str(exc)
    assert str(int) in str(exc)


def test_url(new_suite, suite):
    assert new_suite.url is None
    assert isinstance(suite.url, str)
    assert suite.url == URL



from copy import deepcopy
from datetime import datetime as dt, timedelta

import pytest

from traw.models import Result, Status, User, Test as _Test  # Avoid tox test collection warning


ASSIGNED_TO_ID = 1
COMMENT = "mock result comment"
CREATED_BY_ID = 2
CREATED_ON = 1469726522
DEF1 = "DEFECT1"
DEF2 = "DEFECT2"
DEF3 = "DEFECT3"
ELAPSED = "10m 5s"
ID = 3
STATUS_ID = 4
TEST_ID = 5
VERSION = "1.0RC1"


@pytest.fixture()
def empty_result(client):
    return Result(client)


@pytest.fixture()
def result(client):
    content = {"assignedto_id": ASSIGNED_TO_ID,
               "comment": COMMENT,
               "created_by": CREATED_BY_ID,
               "created_on": CREATED_ON,
               "defects": ",".join([DEF1, DEF2, DEF3]),
               "elapsed": ELAPSED,
               "id": ID,
               "status_id": STATUS_ID,
               "test_id": TEST_ID,
               "version": VERSION}
    return Result(client, deepcopy(content))


def test_assigned_to_get(empty_result, result):
    assert empty_result.assigned_to is None

    result.client.api.user_by_id.return_value = {'id': ASSIGNED_TO_ID}

    assert isinstance(result.assigned_to, User)
    assert result.assigned_to.id == ASSIGNED_TO_ID
    result.client.api.user_by_id.assert_called_with(ASSIGNED_TO_ID)


def test_assigned_to_set(client, empty_result, result):
    user1 = User(client, {'id': ASSIGNED_TO_ID})
    user2 = User(client, {'id': ASSIGNED_TO_ID + 1})

    result.client.api.user_by_id.return_value = {'id': ASSIGNED_TO_ID}

    assert empty_result.assigned_to is None
    empty_result.assigned_to = user1
    assert empty_result._content['assignedto_id'] == ASSIGNED_TO_ID

    assert result.assigned_to.id == ASSIGNED_TO_ID
    result.assigned_to = user2
    assert result._content['assignedto_id'] == ASSIGNED_TO_ID + 1

    result.client.api.user_by_id.assert_called_with(ASSIGNED_TO_ID)


def test_assigned_to_set_exc(client, empty_result):

    assert empty_result.assigned_to is None

    with pytest.raises(TypeError) as exc:
        empty_result.assigned_to = 1234

    assert str(User) in str(exc)
    assert str(int) in str(exc)


def test_comment_get(empty_result, result):
    assert empty_result.comment is None
    assert result.comment == COMMENT


def test_comment_set(empty_result, result):
    empty_result.comment = COMMENT
    assert empty_result.comment == COMMENT

    result.comment = COMMENT + " addition"
    assert result.comment == COMMENT + " addition"


def test_comment_set_exc(empty_result):
    with pytest.raises(TypeError) as exc:
        empty_result.comment = 1234

    assert str(str) in str(exc)
    assert str(int) in str(exc)


def test_created_by(empty_result, result):
    result.client.api.user_by_id.return_value = {'id': CREATED_BY_ID}

    assert empty_result.assigned_to is None
    assert isinstance(result.created_by, User)
    assert result.created_by.id == CREATED_BY_ID

    result.client.api.user_by_id.assert_called_with(CREATED_BY_ID)


def test_created_on(empty_result, result):
    assert empty_result.created_on is None
    assert isinstance(result.created_on, dt)


def test_defects_get(empty_result, result):
    assert list(empty_result.defects) == list()
    assert len(list(result.defects)) == 3
    assert all(map(lambda x: isinstance(x, str), result.defects))
    assert DEF1 in result.defects
    assert DEF2 in result.defects
    assert DEF3 in result.defects


def test_defects_set_str_1(empty_result):
    assert list(empty_result.defects) == list()

    empty_result.defects = str(DEF1)

    assert empty_result._content['defects'] == DEF1


def test_defects_set_str_2(empty_result):
    assert list(empty_result.defects) == list()

    empty_result.defects = ','.join([DEF1, DEF2])

    assert empty_result._content['defects'] == '{0},{1}'.format(DEF1, DEF2)


def test_defects_set_list(empty_result):
    assert list(empty_result.defects) == list()

    empty_result.defects = [DEF1, DEF2]

    assert empty_result._content['defects'] == '{0},{1}'.format(DEF1, DEF2)


def test_defects_set_exc_1(empty_result):
    assert list(empty_result.defects) == list()

    with pytest.raises(TypeError) as exc:
        empty_result.defects = 1234

    assert 'Found 1234' in str(exc)
    assert 'DEFECT1,DEFECT2' in str(exc)


def test_defects_set_exc_2(empty_result):
    assert list(empty_result.defects) == list()

    with pytest.raises(TypeError) as exc:
        empty_result.defects = [1234, 5678]

    assert 'Found [1234, 5678]' in str(exc)
    assert 'DEFECT1,DEFECT2' in str(exc)


def test_elapsed_get(empty_result, result):
    assert empty_result.elapsed is None
    assert isinstance(result.elapsed, timedelta)
    assert result.elapsed == timedelta(seconds=605)


def test_elapsed_set(empty_result):
    assert empty_result.elapsed is None

    empty_result.elapsed = timedelta(seconds=123)

    assert empty_result._content['elapsed'] == 123


def test_elapsed_exc(empty_result):
    assert empty_result.elapsed is None

    with pytest.raises(TypeError) as exc:
        empty_result.elapsed = 'asdf'

    assert str(timedelta) in str(exc)
    assert str(str) in str(exc)


def test_status_get(empty_result, result):
    assert empty_result.status is None

    result.client.api.statuses.return_value = [{'id': STATUS_ID - 1},
                                               {'id': STATUS_ID},
                                               {'id': STATUS_ID + 1}]

    assert isinstance(result.status, Status)
    assert result.status.id == STATUS_ID
    result.client.api.statuses.assert_called_with()


def test_status_set(client, empty_result, result):
    status1 = Status(client, {'id': STATUS_ID})
    status2 = Status(client, {'id': STATUS_ID + 1})

    result.client.api.statuses.return_value = [{'id': STATUS_ID - 1},
                                               {'id': STATUS_ID},
                                               {'id': STATUS_ID + 1}]

    assert empty_result.status is None
    empty_result.status = status1
    assert empty_result._content['status_id'] == STATUS_ID

    assert result.status.id == STATUS_ID
    result.status = status2
    assert result._content['status_id'] == STATUS_ID + 1

    result.client.api.statuses.assert_called_with()


def test_status_set_exc(client, empty_result):

    assert empty_result.status is None

    with pytest.raises(TypeError) as exc:
        empty_result.status = 1234

    assert str(Status) in str(exc)
    assert str(int) in str(exc)


def test_test_get(empty_result, result):
    assert empty_result.test is None

    result.client.api.test_by_id.return_value = {'id': TEST_ID}

    assert isinstance(result.test, _Test)
    assert result.test.id == TEST_ID
    result.client.api.test_by_id.assert_called_with(TEST_ID)


def test_test_set(client, empty_result, result):
    test1 = _Test(client, {'id': TEST_ID})
    test2 = _Test(client, {'id': TEST_ID + 1})

    result.client.api.test_by_id.return_value = {'id': TEST_ID}

    assert empty_result.test is None
    empty_result.test = test1
    assert empty_result._content['test_id'] == TEST_ID

    assert result.test.id == TEST_ID
    result.test = test2
    assert result._content['test_id'] == TEST_ID + 1

    result.client.api.test_by_id.assert_called_with(TEST_ID)


def test_test_set_exc(client, empty_result):

    assert empty_result.test is None

    with pytest.raises(TypeError) as exc:
        empty_result.test = 1234

    assert str(_Test) in str(exc)
    assert str(int) in str(exc)


def test_version_get(empty_result, result):
    assert empty_result.version is None
    assert result.version == VERSION


def test_version_set(empty_result, result):
    empty_result.version = VERSION
    assert empty_result.version == VERSION

    result.version = VERSION + " addition"
    assert result.version == VERSION + " addition"


def test_version_set_exc(empty_result):
    with pytest.raises(TypeError) as exc:
        empty_result.version = 1234

    assert str(str) in str(exc)
    assert str(int) in str(exc)



from .. import const
from .model_base import ModelBase
from .posters import Addable, Deleteable, Updatable
from .project import Project
from .suite import Suite


class Section(Addable, Deleteable, Updatable, ModelBase):
    """ Object model for TestRail Sections

    To add a new section

    .. code-block:: python

        # Get the project and suite you'll be working with
        project = traw_client.project(15)  # Project with ID of 15
        suite = traw_client.suite(12)      # Suite with ID of 12

        # Get an empty section object
        new_section = traw_client.section()
        new_section.name = "New project section"
        new_section.description = "New section description"
        new_section.project = project
        new_section.suite = suite

        # Add it
        section = traw_client.add(new_section)

    """
    _ADDABLE_FIELDS = const.SECTION_ADD_FIELDS
    _UPDATABLE_FIELDS = const.SECTION_UPDATE_FIELDS

    @property
    def depth(self):
        """ The level in the section hierarchy of the test suite """
        return self._content.get('depth', None)

    @property
    def description(self):
        """ The description of the section"""
        return self._content.get('description')

    @description.setter
    def description(self, val):
        if not isinstance(val, str):
            raise TypeError(const.SETTER_ERR.format(str, type(val)))
        self._content['description'] = val

    @property
    def display_order(self):
        """ The order in the test suite """
        return self._content.get('display_order', None)

    @property
    def name(self):
        """ The name of the section"""
        return self._content.get('name')

    @name.setter
    def name(self, val):
        if not isinstance(val, str):
            raise TypeError(const.SETTER_ERR.format(str, type(val)))
        self._content['name'] = val

    @property
    def parent(self):
        """ The parent section in the test suite """
        parent_id = self._content.get('parent_id', None)
        return self.client.section(parent_id) if parent_id is not None else None

    @parent.setter
    def parent(self, val):
        if not isinstance(val, Section):
            raise TypeError(const.SETTER_ERR.format(Section, type(val)))
        self._content['parent_id'] = val.id

    @property
    def project(self):
        """ The project this test section belong to """
        project_id = self._content.get('project_id', None)
        return self.client.project(project_id) if project_id is not None else None

    @project.setter
    def project(self, val):
        if not isinstance(val, Project):
            raise TypeError(const.SETTER_ERR.format(Project, type(val)))
        self._content['project_id'] = val.id

    @property
    def suite(self):
        """ The suite this test section belong to """
        suite_id = self._content.get('suite_id', None)
        return self.client.suite(suite_id) if suite_id is not None else None

    @suite.setter
    def suite(self, val):
        if not isinstance(val, Suite):
            raise TypeError(const.SETTER_ERR.format(Suite, type(val)))
        self._content['suite_id'] = val.id



from copy import deepcopy
from datetime import datetime as dt, timedelta
from functools import update_wrapper, wraps
from inspect import isclass
import re

from singledispatch import singledispatch

from .const import DEFAULT_LIMIT


def cacheable_generator(obj_type):
    """ Caching decorator for API generator methods

        If the decorated method has cached objects for that method and argument
        combination and those objects have not expired, the cached objects are
        yielded without calling the underlying method. If there are no objects in
        that method's cache or the objects have expired, the underlying API method
        is called and the resulting objects are then cached and yielded.

        Cache object expiration is based on the obj_type, and defaults to
        traw.const.DEFAULT_CACHE_TIMEOUT (300 seconds). Cache expiry timeouts can
        be adjusted on a per-object bases from the client:

        .. code-block:: python

            client.change_cache_timeout(models.Run, 30)

        The above will set the cache timeout of models.Run objects from 300
        seconds to 30 seconds
    """
    def _cacheable_generator(func):
        """ """
        cache = func.cache = {}

        @wraps(func)
        def cacheable_func(inst, *args, **kwargs):
            key = str(args) + str(kwargs)
            if key not in cache or cache[key]['expires'] < dt.now():
                returned_vals = list()
                timeout = inst.cache_timeouts[inst][obj_type]
                expires = dt.now() + timedelta(seconds=timeout)
                for val in func(inst, *args, **kwargs):
                    returned_vals.append(val)
                    yield val
                else:
                    # Only cache results if the generator has been exhausted
                    cache[key] = dict()
                    cache[key]['value'] = returned_vals
                    cache[key]['expires'] = expires
            else:
                for val in cache[key]['value']:
                    yield val

        return cacheable_func
    return _cacheable_generator


def cacheable(obj_type):
    """ Caching decorator for API methods that return a single object

        If the decorated method has a cached object for that method and argument
        combination and that object has not expired, the cached object is
        returned without calling the underlying method. If there is no object in
        that method's cache or the object has expired, the underlying API method
        is called and the resulting object is then cached and returned.

        Cache object expiration is based on the obj_type, and defaults to
        traw.const.DEFAULT_CACHE_TIMEOUT (300 seconds). Cache expiry timeouts can
        be adjusted on a per-object bases from the client:

        .. code-block:: python

            client.change_cache_timeout(models.Run, 30)

        The above will set the cache timeout of models.Run objects from 300
        seconds to 30 seconds
    """
    def cacheable_func(func):
        """ """
        cache = func.cache = {}

        @wraps(func)
        def _cacheable_func(inst, *args, **kwargs):
            key = str(args) + str(kwargs)
            if key not in cache or cache[key]['expires'] < dt.now():
                timeout = inst.cache_timeouts[inst][obj_type]
                cache[key] = dict()
                cache[key]['value'] = func(inst, *args, **kwargs)
                cache[key]['expires'] = dt.now() + timedelta(seconds=timeout)

            return cache[key]['value']

        return _cacheable_func
    return cacheable_func


def clear_cache(method):
    """ API method decorator for API methods that POST to the TestRail API

        When TRAW adds/closes/deletes/updates ojects to the TestRail API, any
        objects of the same type that are currently cached by TRAW are
        considered stale. To insure TRAW does not return a cached object that
        has been modified on the TestRail side, ``clear_cache`` is used to
        clear the cache of any API method that works with that same object
        type.

        For instance, if ``traw.api.project_by_id`` has been decorated with the
        ``@cacheable`` decorator, then any API method that POSTs changes to the
        TestRail API for models.Project objects should be decorated with
        ``@clear_cache(project_by_id)``

        .. code-block:: python

            class API(object):

                @cacheable
                def foo_by_id(self, foo_id):
                    # ...
                    return foo

                @cacheable_generator
                def foos(self):
                    # ...
                    yield from foo_list

                @clear_cache(foo_by_id)
                @clear_cache(foos)
                def add_foo(self, new_foo):
                    # ...
                    return new_foo_response

        The method cache is only cleared if the reponse is successful;
        exceptions raises from the actual call to TestRail's API will not clear
        the cache.

    """
    def target(func):
        @wraps(func)
        def _func(*args, **kwargs):
            response = func(*args, **kwargs)
            method.cache.clear()
            return response
        return _func
    return target


def dispatchmethod(func):
    """ singledispatch for class methods

    This provides for a way to use ``functools.singledispatch`` inside of a class.
    It has the same basic interface that ``singledispatch`` does.
    """
    # This implementation builds on the following gist:
    # https://gist.github.com/adamnew123456/9218f99ba35da225ca11
    dispatcher = singledispatch(func)

    def register(type):  # pylint: disable=redefined-builtin
        def _register(func):
            return dispatcher.register(type)(func)

        return _register

    def dispatch(type):  # pylint: disable=redefined-builtin
        return dispatcher.dispatch(type)

    def wrapper(inst, *args, **kwargs):
        obj = args[0] if len(args) > 0 else inst
        cls = obj if isclass(obj) else obj.__class__
        impl = dispatch(cls)
        return impl(inst, *args, **kwargs)

    wrapper.register = register
    wrapper.dispatch = dispatch
    wrapper.registry = dispatcher.registry
    wrapper._clear_cache = dispatcher._clear_cache  # pylint: disable=protected-access
    update_wrapper(wrapper, func)
    return wrapper


def duration_to_timedelta(duration):
    def timespan(segment):
        return int(segment.group(0)[:-1]) if segment else 0

    timedelta_map = {
        'weeks': timespan(re.search('\d+w', duration)),
        'days': timespan(re.search('\d+d', duration)),
        'hours': timespan(re.search('\d+h', duration)),
        'minutes': timespan(re.search('\d+m', duration)),
        'seconds': timespan(re.search('\d+s', duration))
    }
    return timedelta(**timedelta_map)


def paginate(func):
    """ """
    @wraps(func)
    def paginated_func(*args, **kwargs):
        limit = kwargs.get('limit', None)
        offset = 0

        keep_paging = True
        while keep_paging:
            new_kwargs = deepcopy(kwargs)
            new_kwargs['offset'] = offset
            if limit:
                new_kwargs['limit'] = min([limit - offset, DEFAULT_LIMIT])

            obj_count = 0
            for obj_count, obj in enumerate(func(*args, **new_kwargs), 1):
                yield obj

                if limit and obj_count + offset >= limit:
                    break

            offset = offset + obj_count
            if limit and offset >= limit:
                keep_paging = False
            elif limit is None and DEFAULT_LIMIT > obj_count:
                # If obj_count is less than the paging size (DEFAULT_LIMIT),
                # it indicates that there are no more objects for the API
                # to return
                keep_paging = False

    return paginated_func



from datetime import datetime as dt, timedelta

from .. import const
from .case_type import CaseType
from .milestone import Milestone
from .model_base import ModelBase
from .posters import Addable, Deleteable, Updatable
from .priority import Priority
from .section import Section
from .suite import Suite
from .template import Template
from ..utils import duration_to_timedelta as dur_to_td


class Case(Addable, Deleteable, Updatable, ModelBase):
    """ Object model for TestRail Cases

    To create new case

    .. code-block:: python

        new_case = traw_client.case()
        new_case.title = "My new TestRail Case"
        new_case.section = section

        case = client.add(new_case)

    """
    _ADDABLE_FIELDS = const.CASE_ADD_FIELDS
    _UPDATABLE_FIELDS = const.CASE_UPDATE_FIELDS

    @property
    def created_by(self):
        """ The user who created the test case """
        user_id = self._content.get('created_by', None)
        return self.client.user(user_id) if user_id is not None else None

    @property
    def created_on(self):
        """ The date/time when the test case was created
            (as datetime.datetime object)
        """
        co = self._content.get('created_on', None)
        return dt.fromtimestamp(co) if co is not None else None

    @property
    def estimate(self):
        """ The estimated duration of the test case, e.g. "30s" or "1m 45s" """
        est = self._content.get('estimate', None)
        return dur_to_td(est) if est is not None else None

    @estimate.setter
    def estimate(self, val):
        if not isinstance(val, timedelta):
            raise TypeError(const.SETTER_ERR.format(timedelta, type(val)))
        self._content['estimate'] = val.seconds

    @property
    def estimate_forecast(self):
        """ The forecasted, estimated duration of the test case,
            e.g. "30s" or "1m 45s"
        """
        est_forecast = self._content.get('estimate_forecast', None)
        return dur_to_td(est_forecast) if est_forecast is not None else None

    @property
    def milestone(self):  # TODO: find out why API returns only None
        """ The  milestone this test case belongs to """
        milestone_id = self._content.get('milestone_id', None)
        return self.client.milestone(milestone_id) if milestone_id is not None else None

    @milestone.setter
    def milestone(self, val):
        if not isinstance(val, Milestone):
            raise TypeError(const.SETTER_ERR.format(Milestone, type(val)))
        self._content['milestone_id'] = val.id

    @property
    def priority(self):  # TODO: find out why API returns only None
        """ The  priority that is linked to the test case """
        priority_id = self._content.get('priority_id', None)
        return self.client.priority(priority_id) if priority_id is not None else None

    @priority.setter
    def priority(self, val):
        if not isinstance(val, Priority):
            raise TypeError(const.SETTER_ERR.format(Priority, type(val)))
        self._content['priority_id'] = val.id

    @property
    def refs(self):
        """ Yields string references/requirements that are linked to the
            test case
        """
        refs_raw = self._content.get('refs', None)
        refs = refs_raw.split(',') if refs_raw is not None else list()

        for ref in refs:
            yield ref

    @refs.setter
    def refs(self, val):
        """ Add references/requirements associated with the test case

            Refs can be added as follows:
             - A reference string: ``result.refs = "REF01"``
             - A comma separated string: ``result.refs = "REF01,REF02"
             - A list of reference strings: ["REF01", "REF02"]
        """
        if isinstance(val, str):
            refs_ = val
        elif isinstance(val, list) and all(map(lambda x: isinstance(x, str), val)):
            refs_ = ','.join(val)
        else:
            msg = ("References must be either a single string of one or more "
                   "comma separated references ('REF01' or 'REF01,REF02') or "
                   "a list of reference strings (['REF01', 'REF02']). Found {0}")
            raise TypeError(msg.format(val))

        self._content['refs'] = refs_

    @property
    def section(self):
        """ The section that the test case belongs to """
        section_id = self._content.get('section_id', None)
        return self.client.section(section_id) if section_id is not None else None

    @section.setter
    def section(self, val):
        if not isinstance(val, Section):
            raise TypeError(const.SETTER_ERR.format(Section, type(val)))
        self._content['section_id'] = val.id

    @property
    def suite(self):
        """ The suite that the test case belongs to """
        suite_id = self._content.get('suite_id', None)
        return self.client.suite(suite_id) if suite_id is not None else None

    @suite.setter
    def suite(self, val):
        if not isinstance(val, Suite):
            raise TypeError(const.SETTER_ERR.format(Suite, type(val)))
        self._content['suite_id'] = val.id

    @property
    def template(self):
        """ The template that the test case uses """
        template_id = self._content.get('template_id', None)
        return self.client.template(template_id) if template_id is not None else None

    @template.setter
    def template(self, val):
        if not isinstance(val, Template):
            raise TypeError(const.SETTER_ERR.format(Template, type(val)))
        self._content['template_id'] = val.id

    @property
    def title(self):
        """ The title of the test case """
        return self._content.get('title', None)

    @title.setter
    def title(self, val):
        if not isinstance(val, str):
            raise TypeError(const.SETTER_ERR.format(str, type(val)))
        self._content['title'] = val

    @property
    def case_type(self):
        """ The case type that is linked to the test case """
        type_id = self._content.get('type_id', None)
        return self.client.case_type(type_id) if type_id is not None else None

    @case_type.setter
    def case_type(self, val):
        if not isinstance(val, CaseType):
            raise TypeError(const.SETTER_ERR.format(CaseType, type(val)))
        self._content['type_id'] = val.id

    @property
    def updated_by(self):
        """ The user who updated the test case """
        user_id = self._content.get('updated_by', None)
        return self.client.user(user_id) if user_id is not None else None

    @property
    def updated_on(self):
        """ The date/time when the test case was updated
            (as datetime.datetime object)
        """
        co = self._content.get('updated_on', None)
        return dt.fromtimestamp(co) if co is not None else None



class ModelBase(object):
    """ Base class for all TRAW models """
    def __init__(self, client, content=None):
        self.client = client
        self._content = content or dict()

    def __repr__(self):  # pragma: no cover
        return str(self)

    def __str__(self):  # pragma: no cover
        class_name = self.__class__.__name__
        return "{0}-{1}".format(class_name, self.id)

    @property
    def id(self):
        """ The unique ID of the project """
        return self._content.get('id', None)



from types import GeneratorType
from datetime import datetime as dt

from .case import Case
from .. import const
from .milestone import Milestone
from .model_base import ModelBase
from .posters import Addable, Closeable, Deleteable, Updatable
from .project import Project
from .suite import Suite
from .user import User


class Run(Addable, Closeable, Deleteable, Updatable, ModelBase):
    """ Object model for TestRail Runs

    To create new run

    .. code-block:: python

        new_run = traw_client.run()
        new_run.name = "My new TestRail Run"
        new_run.description = "My new run description"

    """
    _ADDABLE_FIELDS = const.RUN_ADD_FIELDS
    _UPDATABLE_FIELDS = const.RUN_UPDATE_FIELDS

    @property
    def add_params(self):
        """ Returns params necessary to add the subclass object to TestRail """
        fields = {field: self._content.get(field, None) for field in self._ADDABLE_FIELDS}

        if fields['case_ids'] is None or fields['case_ids'] == list():
            fields.pop('case_ids')

        return fields

    @property
    def assigned_to(self):
        """ The user the entire test run is assigned to """
        user_id = self._content.get('assignedto_id', None)
        return self.client.user(user_id) if user_id is not None else None

    @assigned_to.setter
    def assigned_to(self, val):
        if not isinstance(val, User):
            raise TypeError(const.SETTER_ERR.format(User, type(val)))
        self._content['assignedto_id'] = val.id

    @property
    def blocked_count(self):
        """ The amount of tests in the test run marked as blocked """
        return self._content.get('blocked_count', None)

    @property
    def cases(self):
        """ A list of cases for the custom case selection """
        case_ids = self._content.get('case_ids', list())
        for case in [self.client.case(c) for c in case_ids]:
            yield case

    @cases.setter
    def cases(self, vals):
        if not isinstance(vals, (list, set, tuple, GeneratorType)):
            msg = ("``cases`` only accepts an iterable (list, set, tuple, "
                   "generator). Found {0}")
            raise TypeError(msg.format(type(vals)))

        case_ids = list()
        for val in vals:
            if not isinstance(val, (Case, int)):
                msg = ("cases iterator can only be models.Case objects or "
                       "int case IDs. Found at least one {0}")
                raise TypeError(msg.format(type(val)))
            case_ids.append(val.id if isinstance(val, Case) else val)
        self._content['case_ids'] = case_ids

    @property
    def completed_on(self):
        """ The date/time when the test run was closed
            (as datetime.datetime object)
        """
        co = self._content.get('completed_on', None)
        return dt.fromtimestamp(co) if co is not None else None

    @property
    def configs(self):
        """ The configs of the test run
            (only present if run is a part of a test plan with configurations)
        """
        if self.project is None:
            raise StopIteration

        for config_id in self._content.get('config_ids', list()):
            yield self.client.config(self.project, config_id)

    @property
    def created_by(self):
        """ The user who created the test run """
        user_id = self._content.get('created_by', None)
        return self.client.user(user_id) if user_id is not None else None

    @property
    def created_on(self):
        """ The date/time when the test run was created
            (as datetime.datetime object)
        """
        co = self._content.get('created_on', None)
        return dt.fromtimestamp(co) if co is not None else None

    # TODO: Add support for custom_statusX_count

    @property
    def description(self):
        """ The description of the run. Can be set to any string """
        return self._content.get('description', None)

    @description.setter
    def description(self, val):
        if not isinstance(val, str):
            raise TypeError(const.SETTER_ERR.format(str, type(val)))
        self._content['description'] = val

    @property
    def failed_count(self):
        """ The amount of tests in the test run marked as failed """
        return self._content.get('failed_count', None)

    @property
    def include_all(self):
        """ True if the test run includes all test cases and false otherwise """
        return self._content.get('include_all', True)

    @include_all.setter
    def include_all(self, val):
        if not isinstance(val, bool):
            raise TypeError(const.SETTER_ERR.format(bool, type(val)))
        self._content['include_all'] = val

    @property
    def is_completed(self):
        """ True if the test run was closed and false otherwise """
        return self._content.get('is_completed', False)

    @property
    def milestone(self):  # TODO: find out why API returns only None
        """ The  milestone this test run belongs to """
        milestone_id = self._content.get('milestone_id', None)
        return self.client.milestone(milestone_id) if milestone_id is not None else None

    @milestone.setter
    def milestone(self, val):
        if not isinstance(val, Milestone):
            raise TypeError(const.SETTER_ERR.format(Milestone, type(val)))
        self._content['milestone_id'] = val.id

    @property
    def name(self):
        """ The name of the test run """
        return self._content.get('name', None)

    @name.setter
    def name(self, val):
        if not isinstance(val, str):
            raise TypeError(const.SETTER_ERR.format(str, type(val)))
        self._content['name'] = val

    @property
    def passed_count(self):
        """ The amount of tests in the test run marked as passed """
        return self._content.get('passed_count', None)

    @property
    def plan(self):
        """ The test plan this test run belongs to """
        plan_id = self._content.get('plan_id', None)
        return self.client.plan(plan_id) if plan_id is not None else None

    @property
    def project(self):
        """ The Project the run belongs to """
        project_id = self._content.get('project_id', None)
        return self.client.project(project_id) if project_id is not None else None

    @project.setter
    def project(self, project):
        if not isinstance(project, Project):
            raise TypeError(const.SETTER_ERR.format(Project, type(project)))
        self._content['project_id'] = project.id

    @property
    def retest_count(self):
        """ The amount of tests in the test run marked as retest """
        return self._content.get('retest_count', None)

    @property
    def suite(self):
        """ The test suite for the test run
            (optional if the project is operating in single suite mode,
            required otherwise)
        """
        suite_id = self._content.get('suite_id', None)
        return self.client.suite(suite_id) if suite_id is not None else None

    @suite.setter
    def suite(self, val):
        if not isinstance(val, Suite):
            raise TypeError(const.SETTER_ERR.format(Suite, type(val)))
        self._content['suite_id'] = val.id

    @property
    def untested_count(self):
        """ The amount of tests in the test run marked as untested """
        return self._content.get('untested_count', None)

    @property
    def update_params(self):
        """ Returns params necessary to update the subclass object in TestRail """
        fields = {field: self._content.get(field, None) for field in self._UPDATABLE_FIELDS}
        if fields['case_ids'] is None or fields['case_ids'] == list():
            fields.pop('case_ids')
        else:
            fields['case_ids'] = ','.join(map(str, fields['case_ids']))

        return fields

    @property
    def url(self):
        """ The address/URL of the run in the user interface """
        return self._content.get('url', None)



from __future__ import print_function

import traw


def main():
    client = traw.Client()  # Credentials loaded through ENV vars

    # Get all priorities
    priorities = client.priorities()  # Returns a models.Priority generator
    priority_1 = next(priorities)
    print(priority_1.name)  # Low
    priority_2 = next(priorities)
    print(priority_2.name)  # Medium
    priority_3 = next(priorities)
    print(priority_3.name)  # High
    priority_4 = next(priorities)
    print(priority_4.name)  # Critical

    # Given a priority ID, you can get the priority object
    priority_from_api = client.priority(1)
    print(priority_from_api.name)  # Low

    # Since you cannot programatically add priorities through the API,
    # attempting to call `client.priority()` will raise an exception rather
    # than returning an empty/new Priority object
    try:
        client.priority()
    except NotImplementedError as exc:
        print(str(exc))  # Not implemented directly. You must pass in int object


if __name__ == "__main__":
    main()



import time

import mock
import pytest

import traw

USER = 'mock username'
PASS = 'mock password'
URL = 'mock url'


# Prevent calls to sleep
class SleepMock(object):
    def __init__(self):
        self.mock_obj = mock.MagicMock()

    def __call__(self, *args):
        self.mock_obj(*args)


time.sleep = SleepMock()


@pytest.fixture()
def client():
    with mock.patch('traw.client.API') as api_mock:
        api_mock.return_value = api_mock

        yield traw.Client(username=USER, password=PASS, url=URL)


@pytest.fixture()
def full_client():
    with mock.patch('traw.api.Session') as sess_mock:
        sess_mock.return_value = sess_mock

        yield traw.Client(username=USER, password=PASS, url=URL)



from copy import deepcopy

import pytest

from traw.models import Template

ID = 1234
NAME = 'mock template name'


@pytest.fixture()
def template(client):
    content = {"id": ID,
               "is_default": True,
               "name": NAME}
    return Template(client, deepcopy(content))


def test_id(template):
    assert template.id == ID


def test_is_default(template):
    assert template.is_default is True


def test_name(template):
    assert template.name == NAME



from .model_base import ModelBase


class User(ModelBase):
    """ Object model for TestRail Users

    To get all users

    .. code-block:: python

        users = list(traw_client.users())

    To get a specific user

    .. code-block:: python

        user_123 = traw_client.user(123)  # Gets user with ID 123 from API
        user_dave = traw_client.user('dave@email.com')  # Gets user with email from API

    """
    @property
    def email(self):
        """ The email address of the user as configured in TestRail """
        return self._content.get('email', None)

    @property
    def is_active(self):
        """ True if the user is active and false otherwise """
        return self._content.get('is_active', None)

    @property
    def name(self):
        """ The full name of the user """
        return self._content.get('name', None)



import datetime
from copy import deepcopy

import pytest

from traw.models import Project

ANNOUNCEMENT = 'mock announcement'
COMPLETED_ON = 1469726522
ID = 123
NAME = 'mock name'
URL = 'mock url'


@pytest.fixture()
def empty_proj(client):
    return Project(client)


@pytest.fixture()
def in_progress_proj(client):
    content = {"announcement": ANNOUNCEMENT,
               "completed_on": None,
               "id": ID,
               "is_completed": False,
               "name": NAME,
               "show_announcement": True,
               "suite_mode": 1,
               "url": URL}
    return Project(client, deepcopy(content))


@pytest.fixture()
def complete_proj(client):
    content = {"announcement": None,
               "completed_on": COMPLETED_ON,
               "id": ID,
               "is_completed": True,
               "name": NAME,
               "show_announcement": False,
               "suite_mode": 2,
               "url": URL}
    return Project(client, content)


def test_announcement_get(empty_proj, in_progress_proj, complete_proj):
    assert empty_proj.announcement is None
    assert in_progress_proj.announcement == ANNOUNCEMENT
    assert complete_proj.announcement is None


def test_announcement_set(empty_proj, in_progress_proj, complete_proj):
    empty_proj.announcement = "New announcement 1"
    assert empty_proj.announcement == "New announcement 1"

    in_progress_proj.announcement = "New announcement 2"
    assert in_progress_proj.announcement == "New announcement 2"

    complete_proj.announcement = "New announcement 3"
    assert complete_proj.announcement == "New announcement 3"


def test_announcement_set_exc(empty_proj):
    with pytest.raises(TypeError) as exc:
        empty_proj.announcement = 1234

    assert str(str) in str(exc)
    assert str(int) in str(exc)


def test_completed_on(empty_proj, in_progress_proj, complete_proj):
    assert empty_proj.completed_on is None
    assert in_progress_proj.completed_on is None
    assert isinstance(complete_proj.completed_on, datetime.datetime)


def test_id(empty_proj, in_progress_proj, complete_proj):
    assert empty_proj.id is None
    assert in_progress_proj.id is ID
    assert complete_proj.id is ID


def test_is_completed(empty_proj, in_progress_proj, complete_proj):
    assert empty_proj.is_completed is False
    assert in_progress_proj.is_completed is False
    assert complete_proj.is_completed is True


def test_is_completed_set(empty_proj, in_progress_proj, complete_proj):
    assert empty_proj.is_completed is False
    empty_proj.is_completed = False
    assert empty_proj.is_completed is False
    empty_proj.is_completed = True
    assert empty_proj.is_completed is True

    assert in_progress_proj.is_completed is False
    in_progress_proj.is_completed = True
    assert in_progress_proj.is_completed is True

    assert in_progress_proj.is_completed is True
    in_progress_proj.is_completed = False
    assert in_progress_proj.is_completed is False


def test_is_completed_set_exc(empty_proj):
    with pytest.raises(TypeError) as exc:
        empty_proj.is_completed = 1234

    assert str(bool) in str(exc)
    assert str(int) in str(exc)


def test_name_get(empty_proj, in_progress_proj, complete_proj):
    assert empty_proj.name is None
    assert in_progress_proj.name is NAME
    assert complete_proj.name is NAME


def test_name_set(empty_proj, in_progress_proj, complete_proj):
    empty_proj.name = NAME
    assert empty_proj.name == NAME

    in_progress_proj.name == NAME
    assert in_progress_proj.name is NAME

    complete_proj.name = NAME
    assert complete_proj.name is NAME


def test_name_set_exc(empty_proj):
    with pytest.raises(TypeError) as exc:
        empty_proj.name = 1234

    assert str(str) in str(exc)
    assert str(int) in str(exc)


def test_show_announcement(empty_proj, in_progress_proj, complete_proj):
    assert empty_proj.show_announcement is False
    assert in_progress_proj.show_announcement is True
    assert complete_proj.show_announcement is False


def test_show_announcement_set(empty_proj, in_progress_proj, complete_proj):
    empty_proj.show_announcement = True
    assert empty_proj.show_announcement is True

    in_progress_proj.show_announcement = False
    assert in_progress_proj.show_announcement is False

    complete_proj.show_announcement = True
    assert complete_proj.show_announcement is True


def test_show_announcement_set_exc(empty_proj):
    with pytest.raises(TypeError) as exc:
        empty_proj.show_announcement = 1234

    assert str(bool) in str(exc)
    assert str(int) in str(exc)


def test_suite_mode_get(empty_proj, in_progress_proj, complete_proj):
    assert empty_proj.suite_mode is None
    assert in_progress_proj.suite_mode is 1
    assert complete_proj.suite_mode is 2


def test_suite_mode_set(empty_proj, in_progress_proj, complete_proj):
    empty_proj.suite_mode = 2
    assert empty_proj.suite_mode is 2

    in_progress_proj.suite_mode = 3
    assert in_progress_proj.suite_mode is 3

    complete_proj.suite_mode = 1
    assert complete_proj.suite_mode is 1


def test_suite_mode_typeerror_exc(empty_proj):
    with pytest.raises(TypeError) as exc:
        empty_proj.suite_mode = 'asdf'

    assert str(str) in str(exc)
    assert str(int) in str(exc)


def test_suite_mode_valueerror_exc(empty_proj):
    with pytest.raises(ValueError) as exc:
        empty_proj.suite_mode = 4

    assert '1, 2, or 3' in str(exc)


def test_url(empty_proj, in_progress_proj, complete_proj):
    assert empty_proj.url is None
    assert in_progress_proj.url is URL
    assert complete_proj.url is URL



from __future__ import print_function

import traw


def main():
    client = traw.Client()  # Credentials loaded through ENV vars

    # Get all available case types:
    case_types = client.case_types()  # Returns a case_types generator
    case_type_1 = next(case_types)
    print(case_type_1.name)  # Acceptance
    print(case_type_1.is_default)  # False
    print(isinstance(case_type_1, traw.models.CaseType))  # True

    case_type_2 = next(case_types)
    print(case_type_2.name)  # Accessibility
    print(case_type_2.is_default)  # False
    print(isinstance(case_type_2, traw.models.CaseType))  # True

    # etc

    # Get a specific case type by case_type_id
    specific_case_type = client.case_type(7)  # Corresponds to the `Other` CT
    print(specific_case_type.name)  # Other
    print(specific_case_type.is_default)  # True

    # Since you cannot programatically add Case Types through the API,
    # attempting to call `client.case_type()` will raise an exception rather
    # than returning an empty/new CaseType object
    try:
        client.case_type()
    except NotImplementedError as exc:
        print(str(exc))  # Not implemented directly. You must pass in int object


if __name__ == "__main__":
    main()



from datetime import timedelta as td
import mock
import pytest

import traw
from traw.const import GET, API_PATH as AP
from traw.utils import dispatchmethod, duration_to_timedelta

MOCK_USERNAME = 'mock username'
MOCK_USER_API_KEY = 'mock user api key'
MOCK_PASSWORD = 'mock password'
MOCK_URL = 'mock url'


class Foo(object):
    def __init__(self):
        self.mock_obj = mock.MagicMock()

    @dispatchmethod
    def method_base(self):
        self.mock_obj.method_base(1)

    @method_base.register(int)
    def method_int(self, var):
        self.mock_obj.method_int(var)

    @method_base.register(str)
    def method_str(self, var):
        self.mock_obj.method_str(var)

    @method_base.register(list)
    def method_list(self, var, *args):
        self.mock_obj.method_list(var + list(args))

    @method_base.register(dict)
    def method_dict(self, var, **kwargs):
        self.mock_obj.method_dict(dict(var, **kwargs))


@pytest.fixture()
def dm():
    yield Foo()


@pytest.fixture()
def dt():
    with mock.patch('traw.utils.dt') as dt_mock:
        yield dt_mock


@pytest.fixture()
def timedelta():
    with mock.patch('traw.utils.timedelta') as timedelta_mock:
        yield timedelta_mock


@pytest.fixture()
def api():
    with mock.patch('traw.api.Session') as Session:
        Session.return_value = Session
        yield traw.api.API(username=MOCK_USERNAME,
                           user_api_key=MOCK_USER_API_KEY,
                           password=MOCK_PASSWORD,
                           url=MOCK_URL)


def test_paginate_w_no_limit(api):
    api.results_by_test_id.cache.clear()
    api._session.request.side_effect = [[1] * 250, [2] * 250, [3] * 50]

    TEST_ID = 1

    results = list(api.results_by_test_id(TEST_ID))

    exp_call_1 = mock.call(method=GET,
                           path=AP['get_results'].format(test_id=TEST_ID),
                           params={'offset': 0})

    exp_call_2 = mock.call(method=GET,
                           path=AP['get_results'].format(test_id=TEST_ID),
                           params={'offset': 250})

    exp_call_3 = mock.call(method=GET,
                           path=AP['get_results'].format(test_id=TEST_ID),
                           params={'offset': 500})

    assert len(results) == 550
    assert results == [1] * 250 + [2] * 250 + [3] * 50
    assert api._session.request.call_args_list == [exp_call_1, exp_call_2, exp_call_3]


def test_paginate_w_limit(api):
    api.results_by_test_id.cache.clear()
    api._session.request.side_effect = [[1] * 250, [2] * 250, [3] * 50]

    TEST_ID = 1

    results = list(api.results_by_test_id(TEST_ID, limit=525))

    exp_call_1 = mock.call(method=GET,
                           path=AP['get_results'].format(test_id=TEST_ID),
                           params={'offset': 0, 'limit': 250})

    exp_call_2 = mock.call(method=GET,
                           path=AP['get_results'].format(test_id=TEST_ID),
                           params={'offset': 250, 'limit': 250})

    exp_call_3 = mock.call(method=GET,
                           path=AP['get_results'].format(test_id=TEST_ID),
                           params={'offset': 500, 'limit': 25})

    assert len(results) == 525
    assert results == [1] * 250 + [2] * 250 + [3] * 25
    assert api._session.request.call_args_list == [exp_call_1, exp_call_2, exp_call_3]


def test_cacheable_caching(timedelta, dt, full_client):
    dt.now.return_value = 1
    timedelta.return_value = 2
    full_client.api.user_by_id.cache.clear()
    for _ in range(20):
        full_client.user(1)

    assert full_client.api._session.call_count == 1


def test_cacheable_generator_caching(timedelta, dt, full_client):
    dt.now.return_value = 1
    timedelta.return_value = 2
    full_client.api.users.cache.clear()
    for _ in range(20):
        list(full_client.users())

    assert full_client.api._session.call_count == 1


def test_cacheable_caching_expire(timedelta, dt, full_client):
    dt.now.side_effect = [0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 9, 9]
    timedelta.return_value = 3
    full_client.api._session.request.side_effect = [{'id': 0}, {'id': 1}, {'id': 2}]
    full_client.api.user_by_id.cache.clear()
    vals = list()
    for _ in range(11):
        vals.append(full_client.user(1).id)

    assert full_client.api._session.request.call_count == 3
    assert vals == [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2]


def test_cacheable_generator_caching_expire(timedelta, dt, full_client):
    dt.now.side_effect = [0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 9, 9]
    timedelta.return_value = 3
    user_dicts = [[{'id': 0}], [{'id': 1}], [{'id': 2}]]
    full_client.api._session.request.side_effect = user_dicts
    full_client.api.users.cache.clear()
    vals = list()
    for _ in range(11):
        vals.append(list(full_client.users())[0].id)

    assert full_client.api._session.request.call_count == 3
    assert vals == [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2]


def test_cacheable_clear_cache(timedelta, dt, full_client):
    dt.now.return_value = 1
    timedelta.return_value = 2
    side_effects = [{'id': 3},    # full_client.milestone(1)
                    [{'id': 4}],  # full_client.milestones(123)
                    {'id': 5},    # full_client.add -> milestone.project.id
                    {'id': 6}]    # full_client.add -> response
    full_client.api._session.request.side_effect = side_effects
    full_client.api.milestone_by_id.cache.clear()
    full_client.api.milestones.cache.clear()

    full_client.milestone(1)
    list(full_client.milestones(123))

    assert len(full_client.api.milestone_by_id.cache) == 1
    assert len(full_client.api.milestones.cache) == 1

    full_client.add(traw.models.Milestone(full_client, {'id': 456, 'project_id': 5}))

    assert len(full_client.api.milestone_by_id.cache) == 0
    assert len(full_client.api.milestones.cache) == 0


def test_dispatchmethod_default(dm):
    """ Verify the base method gets called if you call with an
        unregistered type
    """
    dm.method_base()

    assert dm.mock_obj.method_base.called
    assert not dm.mock_obj.method_int.called
    assert not dm.mock_obj.method_str.called
    assert not dm.mock_obj.method_list.called
    assert not dm.mock_obj.method_dict.called
    dm.mock_obj.method_base.assert_called_once_with(1)


def test_dispatchmethod_int(dm):
    """ Verify the int registered method gets called if you call with an int """
    obj = 1234
    dm.method_base(obj)

    assert not dm.mock_obj.method_base.called
    assert dm.mock_obj.method_int.called
    assert not dm.mock_obj.method_str.called
    assert not dm.mock_obj.method_list.called
    assert not dm.mock_obj.method_dict.called
    dm.mock_obj.method_int.assert_called_once_with(obj)


def test_dispatchmethod_str(dm):
    """ Verify the str registered method gets called if you call with a string """
    obj = 'asdf'
    dm.method_base(obj)

    assert not dm.mock_obj.method_base.called
    assert not dm.mock_obj.method_int.called
    assert dm.mock_obj.method_str.called
    assert not dm.mock_obj.method_list.called
    assert not dm.mock_obj.method_dict.called
    dm.mock_obj.method_str.assert_called_once_with(obj)


def test_dispatchmethod_list_with_args(dm):
    """ Verify args can be passed through """
    obj = ['a', 'b', 'c']
    args = [1, 2, 3]
    dm.method_base(obj, *args)

    assert not dm.mock_obj.method_base.called
    assert not dm.mock_obj.method_int.called
    assert not dm.mock_obj.method_str.called
    assert dm.mock_obj.method_list.called
    assert not dm.mock_obj.method_dict.called
    dm.mock_obj.method_list.assert_called_once_with(obj + args)


def test_dispatchmethod_dict_with_kwargs(dm):
    """ Verify key word args can be passed through """
    obj = dict(a='a', b='b')
    kwargs = dict(one=1, two=2)

    dm.method_base(obj, **kwargs)

    assert not dm.mock_obj.method_base.called
    assert not dm.mock_obj.method_int.called
    assert not dm.mock_obj.method_str.called
    assert not dm.mock_obj.method_list.called
    assert dm.mock_obj.method_dict.called
    dm.mock_obj.method_dict.assert_called_once_with(dict(obj, **kwargs))


def test_duration_to_timedelta():
    """ Verify the duration to timedelta conversion """
    duration = '1w 2d 3h 4m 5s'
    d2td = duration_to_timedelta(duration)

    total_days = 7 + 2
    total_seconds = (3 * 60 * 60) + (4 * 60) + 5

    assert isinstance(d2td, td)
    assert d2td == td(days=total_days, seconds=total_seconds)



from copy import deepcopy

import pytest

from traw.models import Project, Section, Suite

DEPTH = 1
DESCRIPTION = "mock section description"
DISPLAY_ORDER = 2
ID = 3
PARENT_ID = 4
NAME = 'mock name'
PROJECT_ID = 5
SECTION_ID = 6
SUITE_ID = 7


@pytest.fixture()
def new_section(client):
    return Section(client)


@pytest.fixture()
def section(client):
    content = {"depth": DEPTH,
               "description": DESCRIPTION,
               "display_order": DISPLAY_ORDER,
               "id": ID,
               "name": NAME,
               "parent_id": PARENT_ID,
               "project_id": PROJECT_ID,
               "suite_id": SUITE_ID}
    return Section(client, deepcopy(content))


def test_depth(new_section, section):
    assert new_section.depth is None
    assert section.depth == DEPTH


def test_description(new_section, section):
    assert new_section.description is None
    assert section.description == DESCRIPTION


def test_description_set(new_section, section):
    assert new_section.description is None
    new_section.description = "New section Description"
    assert new_section.description == "New section Description"

    assert section.description == DESCRIPTION
    section.description = "New section Description"
    assert section.description == "New section Description"


def test_description_set_exc(new_section):
    with pytest.raises(TypeError) as exc:
        new_section.description = 1234

    assert str(str) in str(exc)
    assert str(int) in str(exc)


def test_display_order(new_section, section):
    assert new_section.display_order is None
    assert section.display_order == DISPLAY_ORDER


def test_name_get(new_section, section):
    assert new_section.name is None
    assert section.name == NAME


def test_name_set(new_section, section):
    assert new_section.name is None
    new_section.name = "New section Name"
    assert new_section.name == "New section Name"

    assert section.name == NAME
    section.name = "New section Name"
    assert section.name == "New section Name"


def test_name_set_exc(new_section):
    with pytest.raises(TypeError) as exc:
        new_section.name = 1234

    assert str(str) in str(exc)
    assert str(int) in str(exc)


def test_parent_get(new_section, section):
    assert new_section.parent is None

    section.client.api.section_by_id.return_value = {'id': PARENT_ID}

    assert isinstance(section.parent, Section)
    assert section.parent.id == PARENT_ID
    section.client.api.section_by_id.assert_called_with(PARENT_ID)


def test_parent_set(new_section):
    assert new_section.parent is None

    new_section.parent = Section(None, {'id': PARENT_ID})

    assert new_section._content['parent_id'] == PARENT_ID


def test_parent_set_exc(new_section):
    assert new_section.parent is None

    with pytest.raises(TypeError) as exc:
        new_section.parent = 1234

    assert str(Section) in str(exc)
    assert str(int) in str(exc)


def test_project_get(new_section, section):
    assert new_section.project is None

    section.client.api.project_by_id.return_value = {'id': PROJECT_ID}

    assert isinstance(section.project, Project)
    assert section.project.id == PROJECT_ID
    section.client.api.project_by_id.assert_called_with(PROJECT_ID)


def test_project_set(new_section):
    assert new_section.project is None

    new_section.project = Project(None, {'id': PROJECT_ID})

    assert new_section._content['project_id'] == PROJECT_ID


def test_project_set_exc(new_section):
    assert new_section.project is None

    with pytest.raises(TypeError) as exc:
        new_section.project = 1234

    assert str(Project) in str(exc)
    assert str(int) in str(exc)


def test_suite_get(new_section, section):
    assert new_section.suite is None

    section.client.api.suite_by_id.return_value = {'id': SUITE_ID}

    assert isinstance(section.suite, Suite)
    assert section.suite.id == SUITE_ID
    section.client.api.suite_by_id.assert_called_with(SUITE_ID)


def test_suite_set(new_section):
    assert new_section.suite is None

    new_section.suite = Suite(None, {'id': SUITE_ID})

    assert new_section._content['suite_id'] == SUITE_ID


def test_suite_set_exc(new_section):
    assert new_section.suite is None

    with pytest.raises(TypeError) as exc:
        new_section.suite = 1234

    assert str(Suite) in str(exc)
    assert str(int) in str(exc)



from datetime import datetime as dt

from .. import const
from .model_base import ModelBase
from .posters import Addable, Deleteable, Updatable
from .project import Project


class Suite(Addable, Deleteable, Updatable, ModelBase):
    _ADDABLE_FIELDS = const.SUITE_ADD_FIELDS
    _UPDATABLE_FIELDS = const.SUITE_UPDATE_FIELDS

    @property
    def completed_on(self):
        """ The date/time when the suite was marked as completed
            (as datetime.datetime object)
        """
        co = self._content.get('completed_on', None)
        return dt.fromtimestamp(co) if co is not None else None

    @property
    def description(self):
        """ The description of the suite. Can be set to any string """
        return self._content.get('description', None)

    @description.setter
    def description(self, val):
        if not isinstance(val, str):
            raise TypeError(const.SETTER_ERR.format(str, type(val)))
        self._content['description'] = val

    @property
    def is_baseline(self):
        """ True if the test suite is a baseline test suite and false otherwise """
        return self._content.get('is_baseline', False)

    @property
    def is_completed(self):
        """ True if the test suite is marked as completed and false otherwise """
        return self._content.get('is_completed', False)

    @property
    def is_master(self):
        """ True if the test suite is a master test suite and false otherwise """
        return self._content.get('is_master', False)

    @property
    def name(self):
        """ The full name of the test suite """
        return self._content.get('name', None)

    @name.setter
    def name(self, val):
        if not isinstance(val, str):
            raise TypeError(const.SETTER_ERR.format(str, type(val)))
        self._content['name'] = val

    @property
    def project(self):
        """ The Project the test suite belongs to """
        project_id = self._content.get('project_id', None)
        return self.client.project(project_id) if project_id is not None else None

    @project.setter
    def project(self, project):
        if not isinstance(project, Project):
            raise TypeError(const.SETTER_ERR.format(Project, type(project)))
        self._content['project_id'] = project.id

    @property
    def url(self):
        """ The address/URL of the test suite in the user interface """
        return self._content.get('url', None)



from copy import deepcopy

import pytest

from traw.models import User

EMAIL = 'user.email@mock.com'
ID = 1234
NAME = 'mock name'


@pytest.fixture()
def user(client):
    content = {"email": EMAIL,
               "id": ID,
               "is_active": True,
               "name": NAME}
    return User(client, deepcopy(content))


def test_email(user):
    assert user.email == EMAIL


def test_id(user):
    assert user.id == ID


def test_is_active(user):
    assert user.is_active is True


def test_name(user):
    assert user.name == NAME



from .case import Case
from .case_type import CaseType
from .config import Config, ConfigGroup
from .milestone import Milestone, SubMilestone
from .plan import Plan
from .priority import Priority
from .project import Project
from .result import Result
from .run import Run
from .section import Section
from .status import Status
from .suite import Suite
from .template import Template
from .test import Test
from .user import User

__all__ = ['Case', 'CaseType', 'Config', 'ConfigGroup', 'Milestone', 'Plan',
           'Priority', 'Project', 'Result', 'Run', 'Section', 'Status',
           'SubMilestone', 'Suite', 'Template', 'Test', 'User']



class Addable(object):
    """  """
    @property
    def add_params(self):
        """ Returns params necessary to add the subclass object to TestRail """
        return {field: self._content.get(field, None) for field in self._ADDABLE_FIELDS}


class Closeable(object):
    """  """
    pass


class Deleteable(object):
    """  """
    pass


class Updatable(object):
    """  """
    @property
    def update_params(self):
        """ Returns params necessary to update the subclass object in TestRail """
        return {field: self._content.get(field, None) for field in self._UPDATABLE_FIELDS}



"""Provide exception classes for the TRAW package."""
import sys


if sys.version_info[0] == 2:
    from urlparse import urlparse  # pragma: no cover
else:
    from urllib.parse import urlparse  # pragma: no cover


class TRAWException(Exception):
    """ Base exception for all TRAW exceptions """
    pass


class TRAWLoginError(TRAWException):
    """ Used with TestRail login related issues """
    pass


class TRAWClientError(TRAWException):
    """ Used for client related errors """
    pass


class UnknownCustomStatusError(TRAWClientError):
    """ Raised when an unknown custom status is encountered (e.g. 'custom_status9' """
    pass


# class RequestException(TRAWException):
#     """Indicate that there was an error with the incomplete HTTP request."""
#
#     def __init__(self, original_exception, request_args, request_kwargs):
#         """Initialize a RequestException instance.
#
#         :param original_exception: The original exception that occurred.
#         :param request_args: The arguments to the request function.
#         :param request_kwargs: The keyword arguments to the request function.
#
#         """
#         self.original_exception = original_exception
#         self.request_args = request_args
#         self.request_kwargs = request_kwargs
#         super(RequestException, self).__init__('error with request {}'
#                                                .format(original_exception))


class ResponseException(TRAWException):
    """Indicate that there was an error with the completed HTTP request."""

    def __init__(self, response):
        """Initialize a RequestException instance.

        :param response: A requests.response instance.

        """
        self.response = response
        super(ResponseException, self).__init__('received {} HTTP response'
                                                .format(response.status_code))


class BadRequest(ResponseException):
    """Indicate invalid parameters for the request."""


class Conflict(ResponseException):
    """Indicate a conflicting change in the target resource."""


class Forbidden(ResponseException):
    """Indicate the authentication is not permitted for the request."""


class InsufficientScope(ResponseException):
    """Indicate that the request requires a different scope."""


class InvalidToken(ResponseException):
    """Indicate that the request used an invalid access token."""


class NotFound(ResponseException):
    """Indicate that the requested URL was not found."""


class RateLimited(ResponseException):
    """Indicate the API rate limit was reached.

    This class adds the attribute ``retry_after``, which is the number of seconds to
    wait before retrying.

    """

    def __init__(self, response):
        """Initialize a Redirect exception instance..

        :param response: A requests.response instance containing a location
        header.

        """
        super(RateLimited, self).__init__(response)
        self.retry_after = int(response.headers['Retry-After'])


class Redirect(ResponseException):
    """Indicate the request resulted in a redirect.

    This class adds the attribute ``path``, which is the path to which the
    response redirects.

    """

    def __init__(self, response):
        """Initialize a Redirect exception instance..

        :param response: A requests.response instance containing a location
        header.

        """
        path = urlparse(response.headers['location']).path
        self.path = path[:-5] if path.endswith('.json') else path
        self.response = response
        TRAWException.__init__(  # pylint: disable=non-parent-init-called
            self, 'Redirect to {}'.format(self.path))


class ServerError(ResponseException):
    """Indicate issues on the server end preventing request fulfillment."""


class TooLarge(ResponseException):
    """Indicate that the request data exceeds the allowed limit."""


class UnknownStatusCode(ResponseException):
    """Indicate that the response status code wasn't recognized."""



import mock
import pytest

import traw
from traw.exceptions import TRAWLoginError
from traw.const import ENVs, GET, POST, API_PATH as AP

MOCK_USERNAME = 'mock username'
MOCK_USER_API_KEY = 'mock user api key'
MOCK_PASSWORD = 'mock password'
MOCK_URL = 'mock url'
CASE1 = {'case': 'case1'}
CASE2 = {'case': 'case2'}
CASE3 = {'case': 'case3'}
CG1 = {'config_group': 'config_group1'}
CG2 = {'config_group': 'config_group2'}
CG3 = {'config_group': 'config_group3'}
CONF1 = {'config': 'config1'}
CONF2 = {'config': 'config2'}
CONF3 = {'config': 'config3'}
CT1 = {'casetype': 'casetype1'}
CT2 = {'casetype': 'casetype2'}
CT3 = {'casetype': 'casetype3'}
MILE1 = {'milestone': 'milestone1'}
MILE2 = {'milestone': 'milestone2'}
MILE3 = {'milestone': 'milestone3'}
PLAN1 = {'plan': 'plan1'}
PLAN2 = {'plan': 'plan2'}
PLAN3 = {'plan': 'plan3'}
PRIO1 = {'priority': 'priority1'}
PRIO2 = {'priority': 'priority2'}
PRIO3 = {'priority': 'priority3'}
PROJ1 = {'project': 'project1'}
PROJ2 = {'project': 'project2'}
PROJ3 = {'project': 'project3'}
RESU1 = {'result': 'result1'}
RESU2 = {'result': 'result2'}
RESU3 = {'result': 'result3'}
RUN1 = {'run': 'run1'}
RUN2 = {'run': 'run2'}
RUN3 = {'run': 'run3'}
SECT1 = {'section': 'section1'}
SECT2 = {'section': 'section2'}
SECT3 = {'section': 'section3'}
STAT1 = {'stat': 'stat1'}
STAT2 = {'stat': 'stat2'}
STAT3 = {'stat': 'stat3'}
SUIT1 = {'suite': 'suite1'}
SUIT2 = {'suite': 'suite2'}
SUIT3 = {'suite': 'suite3'}
TEMP1 = {'temp': 'temp1'}
TEMP2 = {'temp': 'temp2'}
TEMP3 = {'temp': 'temp3'}
TEST1 = {'test': 'test1'}
TEST2 = {'test': 'test2'}
TEST3 = {'test': 'test3'}
USER1 = {'user': 'user1'}
USER2 = {'user': 'user2'}
USER3 = {'user': 'user3'}


@pytest.fixture()
def no_env_vars():
    empty_dict = dict()
    with mock.patch.dict('traw.api.os.environ', empty_dict):
        yield empty_dict


@pytest.fixture()
def env_vars():
    env_vars_dict = {ENVs.USER_KEY: MOCK_USERNAME,
                     ENVs.API_KEY: MOCK_USER_API_KEY,
                     ENVs.PASS_KEY: MOCK_PASSWORD,
                     ENVs.URL_KEY: MOCK_URL}
    with mock.patch.dict('traw.api.os.environ', env_vars_dict):
        yield env_vars_dict


@pytest.fixture()
def env_vars_no_key():
    env_vars_dict = {ENVs.USER_KEY: MOCK_USERNAME,
                     ENVs.PASS_KEY: MOCK_PASSWORD,
                     ENVs.URL_KEY: MOCK_URL}
    with mock.patch.dict('traw.api.os.environ', env_vars_dict):
        yield env_vars_dict


@pytest.fixture()
def no_path_mock():
    with mock.patch('traw.api.os.path') as path_mock:
        path_mock.exists.return_value = False
        yield path_mock


@pytest.fixture()
def config_parser_mock():
    traw_config = {'username': MOCK_USERNAME,
                   'user_api_key': MOCK_USER_API_KEY,
                   'password': MOCK_PASSWORD,
                   'url': MOCK_URL}
    cp_mock = mock.MagicMock()
    cp_mock.return_value = cp_mock
    cp_mock.read.return_value = cp_mock
    cp_mock.__getitem__.return_value = traw_config
    with mock.patch('traw.api.ConfigParser', new_callable=cp_mock):
        with mock.patch('traw.api.os.path.exists'):
            yield traw_config


@pytest.fixture()
def api():
    with mock.patch('traw.api.Session') as Session:
        Session.return_value = Session
        yield traw.api.API(username=MOCK_USERNAME,
                           user_api_key=MOCK_USER_API_KEY,
                           password=MOCK_PASSWORD,
                           url=MOCK_URL)


def test___init___with_caller_supplied_credentials(no_env_vars, no_path_mock):
    """ Verify the credentials can be set through the API init call
        Verify that the user_api_key keyword overrides the password keyword
    """
    api = traw.api.API(username=MOCK_USERNAME,
                       user_api_key=MOCK_USER_API_KEY,
                       password=MOCK_PASSWORD,
                       url=MOCK_URL)

    assert api._session._auth[0] == MOCK_USERNAME
    assert api._session._auth[1] == MOCK_USER_API_KEY
    assert api._session._auth[1] != MOCK_PASSWORD
    assert api._session._url == MOCK_URL


def test___init___with_caller_supplied_password(no_env_vars, no_path_mock):
    """ Verify a password can be used instead of the apk key """
    api = traw.api.API(username=MOCK_USERNAME,
                       password=MOCK_PASSWORD,
                       url=MOCK_URL)

    assert api._session._auth[0] == MOCK_USERNAME
    assert api._session._auth[1] != MOCK_USER_API_KEY
    assert api._session._auth[1] == MOCK_PASSWORD
    assert api._session._url == MOCK_URL


def test___init___with_env_credentials(env_vars, no_path_mock):
    """ Verify the credentials can be set through environment variables
        Verify that the user_api_key keyword overrides the password keyword
    """
    api = traw.api.API()

    assert api._session._auth[0] == MOCK_USERNAME
    assert api._session._auth[1] == MOCK_USER_API_KEY
    assert api._session._auth[1] != MOCK_PASSWORD
    assert api._session._url == MOCK_URL


def test___init___with_env_password(env_vars_no_key, no_path_mock):
    """ Verify the credentials can be set through environment variables
        Verify that the user_api_key keyword overrides the password keyword
    """
    api = traw.api.API()

    assert api._session._auth[0] == MOCK_USERNAME
    assert api._session._auth[1] != MOCK_USER_API_KEY
    assert api._session._auth[1] == MOCK_PASSWORD
    assert api._session._url == MOCK_URL


def test__env_var_exception():
    """ Verify an exception is raised if the wrong value is passed in """
    with pytest.raises(ValueError):
        traw.api._env_var('does not exist')


def test__init__with_config_file(no_env_vars, config_parser_mock):
    """ Verify the credentials can be set a configuration file
        Verify that the user_api_key keyword overrides the password keyword
    """
    api = traw.api.API()

    assert api._session._auth[0] == MOCK_USERNAME
    assert api._session._auth[1] == MOCK_USER_API_KEY
    assert api._session._auth[1] != MOCK_PASSWORD
    assert api._session._url == MOCK_URL


def test__init__with_config_file_w_password(no_env_vars, config_parser_mock):
    """ Verify the credentials can be set a configuration file
        Verify that the password keyword can be used
    """
    config_parser_mock['user_api_key'] = None
    api = traw.api.API()

    assert api._session._auth[0] == MOCK_USERNAME
    assert api._session._auth[1] != MOCK_USER_API_KEY
    assert api._session._auth[1] == MOCK_PASSWORD
    assert api._session._url == MOCK_URL


def test__init__no_credentials_exception(no_env_vars, no_path_mock):
    """ Verify that an exception is raised if no credentials are set """
    with pytest.raises(TRAWLoginError):
        traw.api.API()


def test_case_by_id(api):
    """ Verify the ``case_by_id`` method call """
    CASE_ID = 1234
    api._session.request.return_value = CASE1
    case = api.case_by_id(CASE_ID)

    exp_call = mock.call(method=GET, path=AP['get_case'].format(case_id=CASE_ID))

    assert case == CASE1
    assert isinstance(case, dict)
    assert api._session.request.call_args == exp_call


def test_cases_by_project_id(api):
    """ Verify the ``cases_by_project_id`` method call """
    PROJECT_ID = 1234
    api._session.request.return_value = [CASE1, CASE2]

    case = next(api.cases_by_project_id(PROJECT_ID))

    exp_call = mock.call(method=GET,
                         path=AP['get_cases'].format(project_id=PROJECT_ID),
                         params=dict())

    assert case == CASE1
    assert isinstance(case, dict)
    assert api._session.request.call_args == exp_call


def test_cases_by_project_id_w_params(api):
    """ Verify the ``cases_by_project_id`` method call """
    PROJECT_ID = 1234
    SUITE_ID = 2345
    SECTION_ID = 3456
    params = {'suite_id': SUITE_ID, 'section_id': SECTION_ID}
    api._session.request.return_value = [CASE1, CASE2]

    case = next(api.cases_by_project_id(PROJECT_ID, **params))

    exp_call = mock.call(method=GET,
                         path=AP['get_cases'].format(project_id=PROJECT_ID),
                         params={'suite_id': SUITE_ID, 'section_id': SECTION_ID})

    assert case == CASE1
    assert isinstance(case, dict)
    assert api._session.request.call_args == exp_call


def test_case_types(api):
    """ Verify the ``case_types`` method call """
    api._session.request.return_value = [CT1, CT2, CT3]
    ct_list = list(api.case_types())

    exp_call = mock.call(method=GET, path=AP['get_case_types'])

    assert all(map(lambda c: isinstance(c, dict), ct_list))
    assert len(ct_list) == 3
    assert api._session.request.call_args == exp_call


def test_config_add(api):
    """ Verify the ``config_add`` method call """
    CONFIG_GROUP_ID = 15
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = CONF1
    config = api.config_add(CONFIG_GROUP_ID, params=PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['add_config'].format(config_group_id=CONFIG_GROUP_ID),
                         json=PARAMS)

    assert config == CONF1
    assert isinstance(config, dict)
    assert api._session.request.call_args == exp_call


def test_config_delete(api):
    """ Verify the ``config_delete`` method call """
    CONFIG_ID = 15
    api._session.request.return_value = {}
    config = api.config_delete(CONFIG_ID)

    exp_call = mock.call(method=POST,
                         path=AP['delete_config'].format(config_id=CONFIG_ID))

    assert isinstance(config, dict)
    assert config == dict()
    assert api._session.request.call_args == exp_call


def test_config_update(api):
    """ Verify the ``config_update`` method call """
    CONFIG_ID = 15
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = CONF1
    config = api.config_update(CONFIG_ID, PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['update_config'].format(config_id=CONFIG_ID),
                         json=PARAMS)

    assert config == CONF1
    assert isinstance(config, dict)
    assert api._session.request.call_args == exp_call


def test_config_group_add(api):
    """ Verify the ``config_group_add`` method call """
    PROJECT_ID = 15
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = CG1
    con_grp = api.config_group_add(PROJECT_ID, params=PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['add_config_group'].format(project_id=PROJECT_ID),
                         json=PARAMS)

    assert con_grp == CG1
    assert isinstance(con_grp, dict)
    assert api._session.request.call_args == exp_call


def test_config_group_delete(api):
    """ Verify the ``config_group_delete`` method call """
    CONFIG_GROUP_ID = 15
    api._session.request.return_value = {}
    con_grp = api.config_group_delete(CONFIG_GROUP_ID)

    exp_call = mock.call(method=POST,
                         path=AP['delete_config_group'].format(config_group_id=CONFIG_GROUP_ID))

    assert isinstance(con_grp, dict)
    assert con_grp == dict()
    assert api._session.request.call_args == exp_call


def test_config_group_update(api):
    """ Verify the ``config_group_update`` method call """
    CONFIG_GROUP_ID = 15
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = CG1
    con_grp = api.config_group_update(CONFIG_GROUP_ID, PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['update_config_group'].format(config_group_id=CONFIG_GROUP_ID),
                         json=PARAMS)

    assert con_grp == CG1
    assert isinstance(con_grp, dict)
    assert api._session.request.call_args == exp_call


def test_config_groups(api):
    """ Verify the ``config_groups`` method call """
    PROJECT_ID = 15
    api._session.request.return_value = [CG1, CG2, CG3]
    cg_list = list(api.config_groups(PROJECT_ID))

    exp_call = mock.call(method=GET, path=AP['get_configs'].format(project_id=PROJECT_ID))

    assert all(map(lambda c: isinstance(c, dict), cg_list))
    assert len(cg_list) == 3
    assert api._session.request.call_args == exp_call


def test_milestone_add(api):
    """ Verify the ``milestone_add`` method call """
    PROJECT_ID = 15
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = MILE1
    milestone = api.milestone_add(PROJECT_ID, PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['add_milestone'].format(project_id=PROJECT_ID),
                         json=PARAMS)

    assert milestone == MILE1
    assert isinstance(milestone, dict)
    assert api._session.request.call_args == exp_call


def test_milestone_delete(api):
    """ Verify the ``milestone_delete`` method call """
    MILESTONE_ID = 1234
    api._session.request.return_value = {}
    milestone = api.milestone_delete(MILESTONE_ID)

    exp_call = mock.call(method=POST,
                         path=AP['delete_milestone'].format(milestone_id=MILESTONE_ID))

    assert isinstance(milestone, dict)
    assert milestone == dict()
    assert api._session.request.call_args == exp_call


def test_milestone_update(api):
    """ Verify the ``milestone_update`` method call """
    MILESTONE_ID = 1234
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = MILE1
    milestone = api.milestone_update(MILESTONE_ID, PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['update_milestone'].format(milestone_id=MILESTONE_ID),
                         json=PARAMS)

    assert milestone == MILE1
    assert isinstance(milestone, dict)
    assert api._session.request.call_args == exp_call


def test_milestone_by_id(api):
    """ Verify the ``milestone_by_id`` method call """
    MILESTONE_ID = 1234
    api._session.request.return_value = MILE1
    milestone = api.milestone_by_id(MILESTONE_ID)

    exp_call = mock.call(method=GET,
                         path=AP['get_milestone'].format(milestone_id=MILESTONE_ID))

    assert milestone == MILE1
    assert isinstance(milestone, dict)
    assert api._session.request.call_args == exp_call


def test_milestones_w_defaults(api):
    """ Verify the ``milestones`` method call """
    PROJECT_ID = 15
    api._session.request.return_value = [MILE1, MILE2, MILE3]
    mile_list = list(api.milestones(PROJECT_ID))

    exp_call = mock.call(method=GET,
                         path=AP['get_milestones'].format(project_id=PROJECT_ID),
                         params=None)

    assert all(map(lambda m: isinstance(m, dict), mile_list))
    assert len(mile_list) == 3
    assert api._session.request.call_args == exp_call


def test_milestones_w_params(api):
    """ Verify the ``milestones`` method call with parameters """
    PROJECT_ID = 15
    api._session.request.return_value = [MILE1, MILE2, MILE3]
    mile_list = list(api.milestones(PROJECT_ID, True, False))

    exp_call = mock.call(method=GET,
                         path=AP['get_milestones'].format(project_id=PROJECT_ID),
                         params={'is_completed': 1, 'is_started': 0})

    assert all(map(lambda m: isinstance(m, dict), mile_list))
    assert len(mile_list) == 3
    assert api._session.request.call_args == exp_call


def test_plan_by_id(api):
    """ Verify the ``plan_by_id`` method call """
    PLAN_ID = 1234
    api._session.request.return_value = PLAN1
    plan = api.plan_by_id(PLAN_ID)

    exp_call = mock.call(method=GET, path=AP['get_plan'].format(plan_id=PLAN_ID))

    assert plan == PLAN1
    assert isinstance(plan, dict)
    assert api._session.request.call_args == exp_call


def test_priorities(api):
    """ Verify the ``priorities`` method call """
    api._session.request.return_value = [PRIO1, PRIO2, PRIO3]
    prio_list = list(api.priorities())

    exp_call = mock.call(method=GET, path=AP['get_priorities'])

    assert all(map(lambda p: isinstance(p, dict), prio_list))
    assert len(prio_list) == 3
    assert api._session.request.call_args == exp_call


def test_project_add(api):
    """ Verify the ``project_add`` method call """
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = PROJ1
    project = api.project_add(PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['add_project'],
                         json=PARAMS)

    assert project == PROJ1
    assert isinstance(project, dict)
    assert api._session.request.call_args == exp_call


def test_project_by_id(api):
    """ Verify the project_by_id method can be called """
    PROJ_ID = 1234
    PROJ_DICT = {'proj': 'proj_info'}

    api._session.request.return_value = PROJ_DICT

    proj = api.project_by_id(PROJ_ID)

    assert proj is PROJ_DICT
    assert str(PROJ_ID) in str(api._session.request.call_args)


def test_project_delete(api):
    """ Verify the ``project_delete`` method call """
    PROJECT_ID = 15
    api._session.request.return_value = {}
    project = api.project_delete(PROJECT_ID)

    exp_call = mock.call(method=POST,
                         path=AP['delete_project'].format(project_id=PROJECT_ID))

    assert isinstance(project, dict)
    assert project == dict()
    assert api._session.request.call_args == exp_call


def test_project_update(api):
    """ Verify the ``project_update`` method call """
    PROJECT_ID = 15
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = PROJ1
    project = api.project_update(PROJECT_ID, PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['update_project'].format(project_id=PROJECT_ID),
                         json=PARAMS)

    assert project == PROJ1
    assert isinstance(project, dict)
    assert api._session.request.call_args == exp_call


def test_projects_no_arg(api):
    """ Verify the ``projects`` method call with no args """
    api._session.request.return_value = [PROJ1, PROJ2, PROJ3]
    proj_list = list(api.projects())

    exp_call = mock.call(method=GET, params=None, path=AP['get_projects'])

    assert all(map(lambda p: isinstance(p, dict), proj_list))
    assert len(proj_list) == 3
    assert api._session.request.call_args == exp_call


def test_projects_with_arg(api):
    """ Verify the ``projects`` method call with an arg """
    api._session.request.return_value = [PROJ1, PROJ2, PROJ3]
    ARG1 = 'arg1'
    proj_list = list(api.projects(ARG1))

    exp_call = mock.call(
        method=GET, params={'is_completed': ARG1}, path=AP['get_projects'])

    assert all(map(lambda p: isinstance(p, dict), proj_list))
    assert len(proj_list) == 3
    assert api._session.request.call_args == exp_call


def test_result_add(api):
    """ Verify the ``result_add`` method call """
    TEST_ID = 1155
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = RESU1
    result = api.result_add(TEST_ID, PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['add_result'].format(test_id=TEST_ID),
                         json=PARAMS)

    assert result == RESU1
    assert isinstance(result, dict)
    assert api._session.request.call_args == exp_call


def test_results_by_test_id_no_status_id(api):
    """ Verify the ``results_by_test_id`` method call with no status_id"""
    TEST_ID = 1234
    api._session.request.return_value = [RESU1, RESU2]

    result = next(api.results_by_test_id(TEST_ID))

    exp_call = mock.call(method=GET,
                         path=AP['get_results'].format(test_id=TEST_ID),
                         params={'offset': 0})

    assert result == RESU1
    assert isinstance(result, dict)
    assert api._session.request.call_args == exp_call


def test_results_by_test_id_w_status_id(api):
    """ Verify the ``results_by_run_id`` method call with status_id"""
    TEST_ID = 1234
    api._session.request.return_value = [RESU1, RESU2]

    result = next(api.results_by_test_id(TEST_ID, status_id="111,222"))

    exp_call = mock.call(method=GET,
                         path=AP['get_results'].format(test_id=TEST_ID),
                         params={'offset': 0, 'status_id': '111,222'})

    assert result == RESU1
    assert isinstance(result, dict)
    assert api._session.request.call_args == exp_call


def test_run_add(api):
    """ Verify the ``run_add`` method call """
    PROJECT_ID = 15
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = RUN1
    run = api.run_add(PROJECT_ID, PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['add_run'].format(project_id=PROJECT_ID),
                         json=PARAMS)

    assert run == RUN1
    assert isinstance(run, dict)
    assert api._session.request.call_args == exp_call


def test_run_by_id(api):
    """ Verify the ``run_by_id`` method call """
    RUN_ID = 1234
    api._session.request.return_value = RUN1
    run = api.run_by_id(RUN_ID)

    exp_call = mock.call(method=GET, path=AP['get_run'].format(run_id=RUN_ID))

    assert run == RUN1
    assert isinstance(run, dict)
    assert api._session.request.call_args == exp_call


def test_run_close(api):
    """ Verify the ``run_close`` method call """
    RUN_ID = 1234
    api._session.request.return_value = RUN1
    run = api.run_close(RUN_ID)

    exp_call = mock.call(method=POST,
                         path=AP['close_run'].format(run_id=RUN_ID))

    assert run == RUN1
    assert isinstance(run, dict)
    assert api._session.request.call_args == exp_call


def test_run_delete(api):
    """ Verify the ``milestone_delete`` method call """
    RUN_ID = 1234
    api._session.request.return_value = {}
    run = api.run_delete(RUN_ID)

    exp_call = mock.call(method=POST, path=AP['delete_run'].format(run_id=RUN_ID))

    assert isinstance(run, dict)
    assert run == dict()
    assert api._session.request.call_args == exp_call


def test_run_update(api):
    """ Verify the ``run_update`` method call """
    RUN_ID = 1234
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = RUN1
    run = api.run_update(RUN_ID, PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['update_run'].format(run_id=RUN_ID),
                         json=PARAMS)

    assert run == RUN1
    assert isinstance(run, dict)
    assert api._session.request.call_args == exp_call


def test_runs_by_project_id(api):
    """ Verify the ``runs_by_project_id`` method call """
    PROJECT_ID = 1234
    api._session.request.return_value = [RUN1, RUN2]

    run = next(api.runs_by_project_id(PROJECT_ID))

    exp_call = mock.call(method=GET,
                         params={'offset': 0},
                         path=AP['get_runs'].format(project_id=PROJECT_ID))

    assert run == RUN1
    assert isinstance(run, dict)
    assert api._session.request.call_args == exp_call


def test_runs_by_project_id_w_paginate_and_limit(api):
    """ Verify the ``runs_by_project_id`` pagination and limit """
    PROJECT_ID = 1234
    api._session.request.side_effect = [[RUN1] * 250, [RUN2] * 250]

    runs = list(api.runs_by_project_id(PROJECT_ID, limit=300))

    exp_call_1 = mock.call(method=GET,
                           params={'limit': 250, 'offset': 0},
                           path=AP['get_runs'].format(project_id=PROJECT_ID))

    exp_call_2 = mock.call(method=GET,
                           params={'limit': 50, 'offset': 250},
                           path=AP['get_runs'].format(project_id=PROJECT_ID))

    assert runs == [RUN1] * 250 + [RUN2] * 50
    assert api._session.request.call_args_list == [exp_call_1, exp_call_2]


def test_runs_by_project_id_w_paginate_and_no_limit(api):
    """ Verify the ``runs_by_project_id`` pagination and no limit """
    PROJECT_ID = 1235
    api._session.request.side_effect = [[RUN1] * 250, [RUN2] * 200]

    runs = list(api.runs_by_project_id(PROJECT_ID))

    exp_call_1 = mock.call(method=GET,
                           params={'offset': 0},
                           path=AP['get_runs'].format(project_id=PROJECT_ID))

    exp_call_2 = mock.call(method=GET,
                           params={'offset': 250},
                           path=AP['get_runs'].format(project_id=PROJECT_ID))

    assert runs == [RUN1] * 250 + [RUN2] * 200
    assert api._session.request.call_args_list == [exp_call_1, exp_call_2]


def test_runs_by_project_id_w_paginate_and_no_results(api):
    """ Verify the ``runs_by_project_id`` pagination with no responses returned """
    PROJECT_ID = 1236
    api._session.request.side_effect = [[], ]

    runs = list(api.runs_by_project_id(PROJECT_ID))

    exp_call = mock.call(method=GET,
                         params={'offset': 0},
                         path=AP['get_runs'].format(project_id=PROJECT_ID))

    assert runs == list()
    assert api._session.request.call_args == exp_call


def test_section_add(api):
    """ Verify the ``section_add`` method call """
    PROJECT_ID = 15
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = SECT1
    section = api.section_add(PROJECT_ID, PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['add_section'].format(project_id=PROJECT_ID),
                         json=PARAMS)

    assert section == SECT1
    assert isinstance(section, dict)
    assert api._session.request.call_args == exp_call


def test_section_by_id(api):
    """ Verify the ``section_by_id`` method call """
    SECTION_ID = 1234
    api._session.request.return_value = SECT1
    section = api.section_by_id(SECTION_ID)

    exp_call = mock.call(method=GET,
                         path=AP['get_section'].format(section_id=SECTION_ID))

    assert section == SECT1
    assert isinstance(section, dict)
    assert api._session.request.call_args == exp_call


def test_sections_by_project_id(api):
    """ Verify the ``sections_by_project_id`` method call """
    PROJECT_ID = 1234
    api._session.request.return_value = [SECT1, SECT2]

    section = next(api.sections_by_project_id(PROJECT_ID))

    exp_call = mock.call(method=GET,
                         path=AP['get_sections'].format(project_id=PROJECT_ID),
                         params=dict())

    assert section == SECT1
    assert isinstance(section, dict)
    assert api._session.request.call_args == exp_call


def test_section_delete(api):
    """ Verify the ``section_delete`` method call """
    SECTION_ID = 1234
    api._session.request.return_value = {}
    section = api.section_delete(SECTION_ID)

    exp_call = mock.call(method=POST,
                         path=AP['delete_section'].format(section_id=SECTION_ID))

    assert isinstance(section, dict)
    assert section == dict()
    assert api._session.request.call_args == exp_call


def test_section_update(api):
    """ Verify the ``section_update`` method call """
    SECTION_ID = 1234
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = SECT1
    section = api.section_update(SECTION_ID, PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['update_section'].format(section_id=SECTION_ID),
                         json=PARAMS)

    assert section == SECT1
    assert isinstance(section, dict)
    assert api._session.request.call_args == exp_call


def test_statuses(api):
    """ Verify the ``statuses`` method call """
    api._session.request.return_value = [STAT1, STAT2, STAT3]
    stat_list = list(api.statuses())

    exp_call = mock.call(method=GET, path=AP['get_statuses'])

    assert all(map(lambda s: isinstance(s, dict), stat_list))
    assert len(stat_list) == 3
    assert api._session.request.call_args == exp_call


def test_suite_add(api):
    """ Verify the ``suite_add`` method call """
    PROJECT_ID = 15
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = SUIT1
    suite = api.suite_add(PROJECT_ID, PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['add_suite'].format(project_id=PROJECT_ID),
                         json=PARAMS)

    assert suite == SUIT1
    assert isinstance(suite, dict)
    assert api._session.request.call_args == exp_call


def test_suite_by_id(api):
    """ Verify the ``suite_by_id`` method call """
    SUITE_ID = 1234
    api._session.request.return_value = SUIT1
    suite = api.suite_by_id(SUITE_ID)

    exp_call = mock.call(method=GET, path=AP['get_suite'].format(suite_id=SUITE_ID))

    assert suite == SUIT1
    assert isinstance(suite, dict)
    assert api._session.request.call_args == exp_call


def test_suites_by_project_id(api):
    """ Verify the ``suites_by_project_id`` method call """
    PROJECT_ID = 1234
    api._session.request.return_value = [SUIT1, SUIT2]

    suite = next(api.suites_by_project_id(PROJECT_ID))

    exp_call = mock.call(method=GET, path=AP['get_suites'].format(project_id=PROJECT_ID))

    assert suite == SUIT1
    assert isinstance(suite, dict)
    assert api._session.request.call_args == exp_call


def test_suite_delete(api):
    """ Verify the ``suite_delete`` method call """
    SUITE_ID = 1234
    api._session.request.return_value = {}
    suite = api.suite_delete(SUITE_ID)

    exp_call = mock.call(method=POST,
                         path=AP['delete_suite'].format(suite_id=SUITE_ID))

    assert isinstance(suite, dict)
    assert suite == dict()
    assert api._session.request.call_args == exp_call


def test_suite_update(api):
    """ Verify the ``suite_update`` method call """
    SUITE_ID = 1234
    PARAMS = {'param_key': 'param_value'}
    api._session.request.return_value = SUIT1
    suite = api.suite_update(SUITE_ID, PARAMS)

    exp_call = mock.call(method=POST,
                         path=AP['update_suite'].format(suite_id=SUITE_ID),
                         json=PARAMS)

    assert suite == SUIT1
    assert isinstance(suite, dict)
    assert api._session.request.call_args == exp_call


def test_templates(api):
    """ Verify the ``templates`` method call """
    PROJECT_ID = 15
    api._session.request.return_value = [TEMP1, TEMP2, TEMP3]
    temp_list = list(api.templates(PROJECT_ID))

    exp_call = mock.call(method=GET, path=AP['get_templates'].format(project_id=PROJECT_ID))

    assert all(map(lambda t: isinstance(t, dict), temp_list))
    assert len(temp_list) == 3
    assert api._session.request.call_args == exp_call


def test_test_by_id(api):
    """ Verify the ``test_by_id`` method call """
    TEST_ID = 1234
    api._session.request.return_value = TEST1
    test = api.test_by_id(TEST_ID)

    exp_call = mock.call(method=GET, path=AP['get_test'].format(test_id=TEST_ID))

    assert test == TEST1
    assert isinstance(test, dict)
    assert api._session.request.call_args == exp_call


def test_tests_by_run_id_no_status_id(api):
    """ Verify the ``tests_by_run_id`` method call with no status_id"""
    RUN_ID = 1234
    api._session.request.return_value = [TEST1, TEST2]

    test = next(api.tests_by_run_id(RUN_ID))

    exp_call = mock.call(
        method=GET, path=AP['get_tests'].format(run_id=RUN_ID), params=None)

    assert test == TEST1
    assert isinstance(test, dict)
    assert api._session.request.call_args == exp_call


def test_tests_by_run_id_w_status_id(api):
    """ Verify the ``tests_by_run_id`` method call with status_id"""
    RUN_ID = 1234
    api._session.request.return_value = [TEST1, TEST2]

    test = next(api.tests_by_run_id(RUN_ID, "111,222"))

    exp_call = mock.call(method=GET,
                         path=AP['get_tests'].format(run_id=RUN_ID),
                         params={'status_id': '111,222'})

    assert test == TEST1
    assert isinstance(test, dict)
    assert api._session.request.call_args == exp_call


def test_user_by_email(api):
    """ Verify the user_by_email method can be called """
    USER_EMAIL = 'mock.user@mock.com'
    USER_DICT = {'email': USER_EMAIL}

    api._session.request.return_value = USER_DICT

    user = api.user_by_email(USER_EMAIL)

    assert user is USER_DICT
    assert USER_EMAIL in str(api._session.request.call_args)


def test_user_by_id(api):
    """ Verify the user_by_id method can be called """
    USER_ID = 1234
    USER_DICT = {'id': USER_ID}

    api._session.request.return_value = USER_DICT

    user = api.user_by_id(USER_ID)

    assert user is USER_DICT
    assert str(USER_ID) in str(api._session.request.call_args)


def test_users(api):
    """ Verify the ``users`` method call """
    api._session.request.return_value = [USER1, USER2, USER3]
    user_list = list(api.users())

    exp_call = mock.call(method=GET, path=AP['get_users'])

    assert all(map(lambda u: isinstance(u, dict), user_list))
    assert len(user_list) == 3
    assert api._session.request.call_args == exp_call



from datetime import datetime as dt

import traw


def main():
    client = traw.Client()  # Credentials loaded through ENV vars

    # Locate the project you will be adding milestones to
    project = client.project(15)

    # Create the parent milestone
    new_parent_ms = client.milestone()  # Creates a new, empty milestone object
    new_parent_ms.name = "{0}'s Yearly Test Results".format(project.name)
    new_parent_ms.description = "{0}'s testing results for the year".format(project.name)
    new_parent_ms.start_on = dt(year=2018, month=1, day=1)  # Start on 2018/01/01
    new_parent_ms.due_on = dt(year=2018, month=12, day=31)  # Due on 2018/12/31

    # Associate this milestone with a project
    new_parent_ms.project = project

    # Add the new milestone to TestRail, replacing it with the Milestone returned
    # from the TestRail API.
    parent_ms = client.add(new_parent_ms)
    assert isinstance(parent_ms, traw.models.Milestone)

    # Create a sub milestone for the first quarter of 2018
    new_sub_ms = client.milestone()  # Creates a new, empty milestone object
    new_sub_ms.name = "{0}'s 1st Quarter Test Results".format(project.name)
    new_sub_ms.description = "{0}'s testing results for the first quarter".format(project.name)
    new_sub_ms.start_on = dt(year=2018, month=1, day=1)  # Start on 2018/01/01
    new_sub_ms.due_on = dt(year=2018, month=3, day=31)  # Due on 2018/03/31

    # Associate this milestone with a project
    new_sub_ms.project = project

    # Add the parent milestone, transforming this into a sub_milestone object
    new_sub_ms = new_sub_ms.add_parent(parent_ms)

    # Add the new milestone to TestRail, replacing it with the Milestone returned
    # from the TestRail API.
    sub_ms = client.add(new_sub_ms)
    assert isinstance(sub_ms, traw.models.SubMilestone)

    # The sub_ms's parent should match the one we created
    assert sub_ms.parent.id == parent_ms.id


if __name__ == "__main__":
    main()



import time
from copy import deepcopy
from datetime import datetime as dt

from .. import const
from .model_base import ModelBase
from .posters import Addable, Deleteable, Updatable
from .project import Project


class MilestoneBase(Addable, Deleteable, Updatable, ModelBase):
    _ADDABLE_FIELDS = const.MILESTONE_ADD_FIELDS
    _UPDATABLE_FIELDS = const.MILESTONE_UPDATE_FIELDS

    @property
    def completed_on(self):
        """ The date/time when the milestone was marked as completed
            (as datetime.datetime object)
        """
        co = self._content.get('completed_on', None)
        return dt.fromtimestamp(co) if co is not None else None

    @property
    def description(self):
        """ The description of the milestone. Can be set to any string """
        return self._content.get('description', None)

    @description.setter
    def description(self, val):
        if not isinstance(val, str):
            raise TypeError(const.SETTER_ERR.format(str, type(val)))
        self._content['description'] = val

    @property
    def due_on(self):
        """ The due date of the milestone. Can be set to a datetime.datetime object
            Defaults to None if no due date has been set
        """
        due_on_ = self._content.get('due_on', None)
        return dt.fromtimestamp(due_on_) if due_on_ is not None else None

    @due_on.setter
    def due_on(self, val):
        # TODO: log warning if setting due date to past
        if not isinstance(val, dt):
            raise TypeError(const.SETTER_ERR.format(dt, type(val)))
        self._content['due_on'] = int(time.mktime(val.timetuple()))

    @property
    def is_completed(self):
        """ True if the milestone is marked as completed and false otherwise """
        return self._content.get('is_completed', False)

    @is_completed.setter
    def is_completed(self, val):
        if not isinstance(val, bool):
            raise TypeError(const.SETTER_ERR.format(bool, type(val)))
        self._content['is_completed'] = val

    @property
    def is_started(self):
        """ True if the milestone is marked as started and false otherwise """
        return self._content.get('is_started', False)

    @is_started.setter
    def is_started(self, val):
        if not isinstance(val, bool):
            raise TypeError(const.SETTER_ERR.format(bool, type(val)))
        self._content['is_started'] = val

    @property
    def name(self):
        """ The full name of the milestone """
        return self._content.get('name', None)

    @name.setter
    def name(self, val):
        if not isinstance(val, str):
            raise TypeError(const.SETTER_ERR.format(str, type(val)))
        self._content['name'] = val

    @property
    def project(self):
        """ The Project the milestone belongs to """
        project_id = self._content.get('project_id', None)
        return self.client.project(project_id) if project_id is not None else None

    @project.setter
    def project(self, project):
        if not isinstance(project, Project):
            raise TypeError(const.SETTER_ERR.format(Project, type(project)))
        self._content['project_id'] = project.id

    @property
    def start_on(self):
        """ The scheduled start date/time of the milestone
            (as datetime.datetime object)
        """
        start_on_ = self._content.get('start_on', None)
        return dt.fromtimestamp(start_on_) if start_on_ is not None else None

    @start_on.setter
    def start_on(self, val):
        if not isinstance(val, dt):
            raise TypeError(const.SETTER_ERR.format(dt, type(val)))
        self._content['start_on'] = int(time.mktime(val.timetuple()))

    @property
    def started_on(self):
        """ The date/time when the milestone was started
            (as datetime.datetime object)
        """
        started_on_ = self._content.get('started_on', None)
        return dt.fromtimestamp(started_on_) if started_on_ is not None else None

    @started_on.setter
    def started_on(self, val):
        if not isinstance(val, dt):
            raise TypeError(const.SETTER_ERR.format(dt, type(val)))
        self._content['started_on'] = int(time.mktime(val.timetuple()))

    @property
    def url(self):
        """ The address/URL of the milestone in the user interface """
        return self._content.get('url', None)


class SubMilestone(MilestoneBase):
    @property
    def parent(self):
        """ The ID of the parent milestone the sub-milestone belongs to

            When assigning a new parent milestone to a sub-milestone, they most both have
            the same project set
        """
        parent_id = self._content.get('parent_id', None)
        return self.client.milestone(parent_id) if parent_id is not None else None

    @parent.setter
    def parent(self, parent_ms):
        if not isinstance(parent_ms, Milestone):
            raise TypeError(const.SETTER_ERR.format(Milestone, type(parent_ms)))
        elif self.project is None:
            msg = "You must set the sub-milestone's project to an existing project first"
            raise ValueError(msg)
        elif self.project.id != parent_ms.project.id:
            msg = ("The sub-milestone and parent milestone project IDs must match. Currently, "
                   "the sub-milestone project ID is {0} and the parent milestone project "
                   "ID is {1}")
            raise ValueError(msg.format(self.project.id, parent_ms.project.id))
        else:
            self._content['parent_id'] = parent_ms.id


class Milestone(MilestoneBase):
    """ Object model for TestRail Milestones

    To get all Milestones

    .. code-block:: python

        project = traw_client.project(1234)  # Get Project with ID 1234
        milestones = list(traw_client.milestones(project))  # Get all milestones for project

    To get a specific milestone

    .. code-block:: python

        milestone_123 = traw_client.milestone(123)  # Gets milestone with ID 123 from API

    To create new project

    .. code-block:: python

        from datetime import timedelta, datetime as dt

        project = traw_client.project(15)  # Get project with project_id = 15

        new_milestone = traw_client.milestone()
        new_milestone.name = "My new TestRail milestone"
        new_milestone.description = "My new milestone description"
        new_milestone.due_on = dt.now() + timedelta(days=14)  # Due in 14 days
        new_milestone.is_completed = False
        new_milestone.project = project

    """
    def add_parent(self, milestone):
        """ Adding a parent milestone to a Milestone object transforms it into a
            SubMilestone

            :param milestone: Milestone or int of the paraent milestone

            :returns: new SubMilestone
        """
        if not isinstance(milestone, (int, Milestone)):
            exp_types = "{0} or {1}".format(int, Milestone)
            raise TypeError(const.SETTER_ERR.format(exp_types, type(milestone)))

        parent_id = milestone if isinstance(milestone, int) else milestone.id

        sub_milestone = deepcopy(self._content)
        sub_milestone['parent_id'] = parent_id

        return SubMilestone(self.client, sub_milestone)

    @property
    def sub_milestones(self):
        """ Returns sub Milestone generator

            The sub-milestones that belong to the milestone (if any)
        """
        # This one is a bit tricky, as sub milestones are only included in the API response
        # if the milestone came from the ``get_milestone`` endpoint. If it came from, for
        # example, ``get_milestones``, the keyword will be completely absent
        sub_milestones_ = self._content.get('milestones', None)
        if sub_milestones_ is None:
            # None indicates this milestone came from somewhere other than the `get_milestone`
            # endpoint. Need to make an additional API call to get the sub milestones
            self._content['milestones'] = self.client.api.milestone_by_id(self.id)['milestones']

        for sub_milestone in self._content['milestones']:
            yield SubMilestone(self.client, sub_milestone)



from copy import deepcopy
from datetime import datetime as dt

import pytest

from traw.models import Milestone, Project, SubMilestone

COMPLETED_ON = 1469726522
DESCRIPTION = "mock milestone description"
DUE_ON = 1469726333
ID = 123
NAME = 'mock name'
PARENT_ID = 1234
PROJECT_ID = 15
START_ON = 1469726444
STARTED_ON = 1469726555
URL = 'mock url'


@pytest.fixture()
def new_milestone(client):
    return Milestone(client)


@pytest.fixture()
def milestone(client):
    content = {"completed_on": COMPLETED_ON,
               "description": DESCRIPTION,
               "due_on": DUE_ON,
               "id": ID,
               "is_completed": False,
               "is_started": False,
               "name": NAME,
               "project_id": PROJECT_ID,
               "start_on": START_ON,
               "started_on": STARTED_ON,
               "url": URL}
    return Milestone(client, deepcopy(content))


@pytest.fixture()
def sub_milestone(client):
    content = {"completed_on": COMPLETED_ON,
               "description": DESCRIPTION,
               "due_on": DUE_ON,
               "id": ID,
               "is_completed": False,
               "is_started": False,
               "is_completed": False,
               "name": NAME,
               "parent_id": PARENT_ID,
               "project_id": PROJECT_ID,
               "start_on": START_ON,
               "started_on": STARTED_ON,
               "url": URL}
    return SubMilestone(client, deepcopy(content))


def test_completed_on(new_milestone, milestone):
    assert new_milestone.completed_on is None
    assert isinstance(milestone.completed_on, dt)
    assert milestone.completed_on == dt.fromtimestamp(COMPLETED_ON)


def test_description_get(new_milestone, milestone):
    assert new_milestone.description is None
    assert milestone.description == DESCRIPTION


def test_description_set(new_milestone, milestone):
    new_milestone.description = "New description 1"
    assert new_milestone.description == "New description 1"

    milestone.description = "New description 2"
    assert milestone.description == "New description 2"


def test_description_set_exc(new_milestone):
    with pytest.raises(TypeError) as exc:
        new_milestone.description = 1234

    assert str(str) in str(exc)
    assert str(int) in str(exc)


def test_due_on_get(new_milestone, milestone):
    assert new_milestone.due_on is None
    assert isinstance(milestone.due_on, dt)
    assert milestone.due_on == dt.fromtimestamp(DUE_ON)


def test_due_on_set(new_milestone):
    assert new_milestone.due_on is None

    new_milestone.due_on = dt.fromtimestamp(DUE_ON)

    assert isinstance(new_milestone.due_on, dt)
    assert new_milestone.due_on == dt.fromtimestamp(DUE_ON)


def test_due_on_set_exc(new_milestone):
    with pytest.raises(TypeError) as exc:
        new_milestone.due_on = 1234

    assert str(dt) in str(exc)
    assert str(int) in str(exc)


def test_is_completed_get(new_milestone, milestone):
    assert new_milestone.is_completed is False
    assert milestone.is_completed is False


def test_is_completed_set(milestone):
    assert milestone.is_completed is False

    milestone.is_completed = True

    assert milestone.is_completed is True


def test_is_completed_set_exc(new_milestone):
    with pytest.raises(TypeError) as exc:
        new_milestone.is_completed = 1234

    assert str(bool) in str(exc)
    assert str(int) in str(exc)


def test_is_started_get(new_milestone, milestone):
    assert new_milestone.is_started is False
    assert milestone.is_started is False


def test_is_started_set(milestone):
    assert milestone.is_started is False

    milestone.is_started = True

    assert milestone.is_started is True


def test_is_started_set_exc(new_milestone):
    with pytest.raises(TypeError) as exc:
        new_milestone.is_started = 1234

    assert str(bool) in str(exc)
    assert str(int) in str(exc)


def test_name_get(new_milestone, milestone):
    assert new_milestone.name is None
    assert milestone.name is NAME


def test_name_set(new_milestone, milestone):
    new_milestone.name = NAME
    assert new_milestone.name == NAME

    milestone.name = NAME
    assert milestone.name == NAME


def test_name_set_exc(new_milestone):
    with pytest.raises(TypeError) as exc:
        new_milestone.name = 1234

    assert str(str) in str(exc)
    assert str(int) in str(exc)


def test_project_get(new_milestone, milestone):
    assert new_milestone.project is None
    assert isinstance(milestone.project, Project)

    milestone.client.api.project_by_id.return_value = {'id': PROJECT_ID}

    assert milestone.project.id == PROJECT_ID
    milestone.client.api.project_by_id.assert_called_with(PROJECT_ID)


def test_project_set(new_milestone):
    project = Project(None, {'id': PROJECT_ID})
    new_milestone.project = project

    assert new_milestone._content['project_id'] == PROJECT_ID


def test_project_set_exc(new_milestone):
    with pytest.raises(TypeError) as exc:
        new_milestone.project = 1234

    assert str(Project) in str(exc)
    assert str(int) in str(exc)


def test_start_on_get(new_milestone, milestone):
    assert new_milestone.start_on is None
    assert isinstance(milestone.start_on, dt)
    assert milestone.start_on == dt.fromtimestamp(START_ON)


def test_start_on_set(new_milestone):
    assert new_milestone.start_on is None

    new_milestone.start_on = dt.fromtimestamp(START_ON)

    assert isinstance(new_milestone.start_on, dt)
    assert new_milestone.start_on == dt.fromtimestamp(START_ON)


def test_start_on_set_exc(new_milestone):
    with pytest.raises(TypeError) as exc:
        new_milestone.start_on = 1234

    assert str(dt) in str(exc)
    assert str(int) in str(exc)


def test_started_on_get(new_milestone, milestone):
    assert new_milestone.started_on is None
    assert isinstance(milestone.started_on, dt)
    assert milestone.started_on == dt.fromtimestamp(STARTED_ON)


def test_started_on_set(new_milestone):
    assert new_milestone.started_on is None

    new_milestone.started_on = dt.fromtimestamp(STARTED_ON)

    assert isinstance(new_milestone.started_on, dt)
    assert new_milestone.started_on == dt.fromtimestamp(STARTED_ON)


def test_started_on_set_exc(new_milestone):
    with pytest.raises(TypeError) as exc:
        new_milestone.started_on = 1234

    assert str(dt) in str(exc)
    assert str(int) in str(exc)


def test_url(new_milestone, milestone):
    assert new_milestone.url is None
    assert isinstance(milestone.url, str)
    assert milestone.url == URL


def test_add_parent_milestone_by_id(new_milestone):
    sub_milestone = new_milestone.add_parent(PARENT_ID)

    assert isinstance(sub_milestone, SubMilestone)
    assert sub_milestone._content['parent_id'] == PARENT_ID


def test_add_parent_milestone_by_milestone(new_milestone, milestone):
    sub_milestone = new_milestone.add_parent(milestone)

    assert isinstance(sub_milestone, SubMilestone)
    assert sub_milestone._content['parent_id'] == milestone.id


def test_add_parent_exc(new_milestone):
    with pytest.raises(TypeError) as exc:
        new_milestone.add_parent('asdf')

    assert str(int) in str(exc)
    assert str(str) in str(exc)
    assert str(Milestone) in str(exc)


def test_sub_milestone_parent_get(sub_milestone):
    sub_milestone.client.api.milestone_by_id.return_value = {'id': PARENT_ID}

    assert isinstance(sub_milestone.parent, Milestone)
    assert sub_milestone.parent.id == PARENT_ID
    sub_milestone.client.api.milestone_by_id.assert_called_with(PARENT_ID)


def test_sub_milestone_parent_set_type_error_exc(sub_milestone):
    with pytest.raises(TypeError) as exc:
        sub_milestone.parent = 1234

    assert str(Milestone) in str(exc)
    assert str(int) in str(exc)


def test_sub_milestone_parent_set_value_error_exc(sub_milestone, milestone):
    sub_milestone._content['project_id'] = None
    with pytest.raises(ValueError) as exc:
        sub_milestone.parent = milestone

    assert 'existing project first' in str(exc)


def test_sub_milestone_parent_set_value_error_project_mismatch_exc(sub_milestone, milestone):
    sub_milestone.client.api.project_by_id.side_effect = [
        {'id': 12345}, {'id': 12345}, {'id': PROJECT_ID}, {'id': 12345}, {'id': PROJECT_ID}]
    sub_milestone._content['project_id'] = 12345
    with pytest.raises(ValueError) as exc:
        sub_milestone.parent = milestone

    assert str(12345) in str(exc)
    assert str(PROJECT_ID) in str(exc)


def test_sub_milestone_parent_set(sub_milestone, milestone):
    milestone._content['id'] = 9876
    sub_milestone.client.api.project_by_id.side_effect = [
        {'id': PROJECT_ID}, {'id': PROJECT_ID}, {'id': PROJECT_ID}]
    sub_milestone.parent = milestone

    assert sub_milestone._content['parent_id'] == 9876


def test_sub_milestones_no_extra_api_call(milestone):
    milestone._content['milestones'] = [{'id': 333}, {'id': 444}]

    assert all([isinstance(sub, SubMilestone) for sub in milestone.sub_milestones])
    assert len(list(milestone.sub_milestones)) == 2
    assert not milestone.client.api.milestone_by_id.called


def test_sub_milestones_with_api_call(milestone):
    milestone.client.api.milestone_by_id.side_effect = [{'milestones': [{'id': 333}, {'id': 444}]}]

    assert all([isinstance(sub, SubMilestone) for sub in milestone.sub_milestones])
    assert len(list(milestone.sub_milestones)) == 2
    milestone.client.api.milestone_by_id.assert_called_once_with(milestone.id)



import mock
import pytest
import requests
from requests.status_codes import codes
import time

from traw.sessions import Session
from traw import exceptions
from traw.const import GET, BASE_API_PATH as BAP, API_PATH as AP

UNAME = 'mock username'
PWORD = 'mock password'
AUTH = (UNAME, PWORD)
URL = 'http://mock.url'


@pytest.fixture()
def session():
    with mock.patch('traw.sessions.requests') as req_mock:
        sess_mock = mock.create_autospec(requests.Session)
        sess_mock.headers = dict()
        req_mock.Session.return_value = sess_mock

        yield Session(auth=AUTH, url=URL)


@pytest.fixture()
def response():
    yield mock.create_autospec(requests.models.Response)


def test___init__():
    """ Validate the session __init__ """
    with mock.patch('traw.sessions.requests') as req_mock:
        sess_mock = mock.create_autospec(requests.Session)
        sess_mock.headers = dict()
        req_mock.Session.return_value = sess_mock
        session = Session(auth=AUTH, url=URL)

        assert session._auth == AUTH
        assert session._url == URL
        assert session._http is sess_mock
        assert session._http.headers['Content-Type'] == 'application/json'


@mock.patch.object(Session, '_request_with_retries')
def test_request_with_defaults(req_mock, session):
    """ Validate the ``session.request`` method call """
    REQ_RESP = {'key': 'value'}
    req_mock.return_value = REQ_RESP
    response = session.request(method=GET, path=AP['get_projects'])

    exp_url = URL + BAP + '/' + AP['get_projects']
    exp_call = mock.call(method=GET, params={}, url=exp_url, json=None)

    assert response is REQ_RESP
    assert req_mock.call_args == exp_call


@mock.patch.object(Session, '_request_with_retries')
def test_request_with_params(req_mock, session):
    """ Validate the ``session.request`` method call with params"""
    REQ_RESP = {'key': 'value'}
    req_mock.return_value = REQ_RESP
    PARAM = {'param_key': 'param_val'}
    JSON = {'json_key': 'json_val'}
    response = session.request(
        method=GET, path=AP['get_projects'], params=PARAM, json=JSON)

    exp_url = URL + BAP + '/' + AP['get_projects']
    exp_call = mock.call(method=GET, params=PARAM, url=exp_url, json=JSON)

    assert response is REQ_RESP
    assert req_mock.call_args == exp_call


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_bad_gateway(make_req_mock, session, response):
    """ Validate _request_with_retries exception logic for bad_gateway """
    response.status_code = codes['bad_gateway']
    make_req_mock.return_value = response, None

    with pytest.raises(exceptions.ServerError):
        session._request_with_retries()

    assert make_req_mock.call_count == 3


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_rate_limited(make_req_mock, session, response):
    """ Validate _request_with_retries exception logic for bad_gateway """
    time.sleep.mock_obj.reset_mock()
    response.status_code = 429
    response.headers = {'Retry-After': 60}
    make_req_mock.return_value = response, None

    with pytest.raises(exceptions.RateLimited):
        session._request_with_retries()

    assert make_req_mock.call_count == 3
    assert str(time.sleep.mock_obj.call_args_list).count('60') == 3


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_bad_request(make_req_mock, session, response):
    """ Validate _request_with_retries exception logic for bad_request """
    response.status_code = codes['bad_request']
    make_req_mock.return_value = response, None

    with pytest.raises(exceptions.BadRequest):
        session._request_with_retries()

    assert make_req_mock.call_count == 1


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_conflict(make_req_mock, session, response):
    """ Validate _request_with_retries exception logic for bad_request """
    response.status_code = codes['conflict']
    make_req_mock.return_value = response, None

    with pytest.raises(exceptions.Conflict):
        session._request_with_retries()

    assert make_req_mock.call_count == 1


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_found(make_req_mock, session):
    """ Validate _request_with_retries exception logic for bad_request """
    MOCK_URL = 'http://mock.domain/mock_loc'
    MOCK_LOC = '/mock_loc'
    response = mock.MagicMock()
    response.status_code = codes['found']
    response.headers = dict()
    response.headers['location'] = MOCK_URL
    make_req_mock.return_value = response, None

    with pytest.raises(exceptions.Redirect) as exc:
        session._request_with_retries()

    assert make_req_mock.call_count == 1
    assert exc.value.path == MOCK_LOC


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_forbidden(make_req_mock, session, response):
    """ Validate _request_with_retries exception logic for forbidden """
    response.status_code = codes['forbidden']
    make_req_mock.return_value = response, None

    with pytest.raises(exceptions.Forbidden):
        session._request_with_retries()

    assert make_req_mock.call_count == 1


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_gateway_timeout(make_req_mock, session, response):
    """ Validate _request_with_retries exception logic for gateway_timeout """
    response.status_code = codes['gateway_timeout']
    make_req_mock.return_value = response, None

    with pytest.raises(exceptions.ServerError):
        session._request_with_retries()

    assert make_req_mock.call_count == 3


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_internal_serv_err(make_req_mock, session, response):
    """ Validate _request_with_retries exception logic for internal_server_error """
    response.status_code = codes['internal_server_error']
    make_req_mock.return_value = response, None

    with pytest.raises(exceptions.ServerError):
        session._request_with_retries()

    assert make_req_mock.call_count == 3


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_not_found(make_req_mock, session, response):
    """ Validate _request_with_retries exception logic for not_found """
    response.status_code = codes['not_found']
    make_req_mock.return_value = response, None

    with pytest.raises(exceptions.NotFound):
        session._request_with_retries()

    assert make_req_mock.call_count == 1


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_req_too_large(make_req_mock, session, response):
    """ Validate _request_with_retries exception logic for req_too_large """
    response.status_code = codes['request_entity_too_large']
    make_req_mock.return_value = response, None

    with pytest.raises(exceptions.TooLarge):
        session._request_with_retries()

    assert make_req_mock.call_count == 1


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_service_unavailable(make_req_mock, session, response):
    """ Validate _request_with_retries exception logic for service_unavailable """
    response.status_code = codes['service_unavailable']
    make_req_mock.return_value = response, None

    with pytest.raises(exceptions.ServerError):
        session._request_with_retries()

    assert make_req_mock.call_count == 3


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_unauthorized(make_req_mock, session, response):
    """ Validate _request_with_retries exception logic for unauthorized """
    response.status_code = codes['unauthorized']
    make_req_mock.return_value = response, None

    with pytest.raises(exceptions.Forbidden):
        session._request_with_retries()

    assert make_req_mock.call_count == 1


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_no_content(make_req_mock, session, response):
    """ Validate _request_with_retries logic for no_content """
    response.status_code = codes['no_content']
    make_req_mock.return_value = response, None

    result = session._request_with_retries()

    assert result is None
    assert make_req_mock.call_count == 1


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_unknown_status_code(make_req_mock, session, response):
    """ Validate _request_with_retries exception logic for unknown status code """
    response.status_code = 666
    make_req_mock.return_value = response, None

    with pytest.raises(exceptions.UnknownStatusCode):
        session._request_with_retries()

    assert make_req_mock.call_count == 1


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_zero_len_header(make_req_mock, session):
    """ Validate _request_with_retries logic for zero length response """
    response = mock.MagicMock()
    response.status_code = codes['ok']
    response.headers = dict()
    response.headers['content-length'] = '0'
    make_req_mock.return_value = response, None

    result = session._request_with_retries()

    assert result == ''
    assert make_req_mock.call_count == 1


@mock.patch.object(Session, '_make_request')
def test_req_w_retries_valid_content(make_req_mock, session):
    """ Validate _request_with_retries exception logic for valid content """
    exp_dict = {'key': 'value'}
    response = mock.MagicMock()
    response.status_code = codes['ok']
    response.headers = dict()
    response.headers['content-length'] = '100'
    response.json.return_value = exp_dict
    make_req_mock.return_value = response, None

    result = session._request_with_retries()

    assert result == exp_dict
    assert make_req_mock.call_count == 1


def test__make_request_exc(session):
    """ Validate _make_request handles exceptions """
    session._http.request.side_effect = ValueError
    resp, exc = session._make_request(method='method', url='url')

    assert resp is None
    assert isinstance(exc, ValueError)


def test__make_request_valid_response(session, response):
    """ Validate _make_request returns response """
    response.status_code = 200
    response.headers = {'content-length': 300}
    session._http.request.return_value = response

    resp, exc = session._make_request(method='method', url='url')

    assert resp is response
    assert exc is None


def test_close(session):
    """ Validate close call """
    session.close()

    assert session._http.close.called



from .model_base import ModelBase


class Priority(ModelBase):
    """ Object model for TestRail Priorities

    To get priorities

    .. code-block:: python

        priorities = list(traw_client.priorities())

    """
    @property
    def is_default(self):
        """ Reports if priority is default """
        return self._content.get('is_default')

    @property
    def name(self):
        """ The (long) name of the priority: `1 - Don't Test`"""
        return self._content.get('name')

    @property
    def priority(self):
        """ The priority number """
        return self._content.get('priority')

    @property
    def short_name(self):
        """ The short name of the priority: `1 - Don't` """
        return self._content.get('short_name')



from copy import deepcopy
import datetime

from mock import call
import pytest

from traw.models import Case, Config, Milestone, Plan, Project, Run, Suite, User

CONF1_ID = 3
CONF2_ID = 4
CG1_ID = 17

ASSIGNED_TO_ID = 1
BLOCKED_COUNT = 2
COMPLETED_ON = 1469726522
CONFIGS = [CONF1_ID, CONF2_ID]
CREATED_BY_ID = 5
CREATED_ON = 1469725522
DESCRIPTION = "mock run description"
FAILED_COUNT = 6
ID = 7
MILESTONE_ID = 8
NAME = 'mock name'
PASSED_COUNT = 9
PLAN_ID = 10
PROJECT_ID = 11
RETEST_COUNT = 12
SUITE_ID = 13
UNTESTED_COUNT = 14
URL = 'mock url'

CASE1_ID = 15
CASE2_ID = 16
CASE1 = {'id': CASE1_ID}
CASE2 = {'id': CASE2_ID}

CONF3 = {'id': 3}
CONF4 = {'id': 4}

CG1 = {'id': CG1_ID, 'configs': [CONF3, CONF4]}


@pytest.fixture()
def new_run(client):
    return Run(client)


@pytest.fixture()
def run(client):
    content = {'assignedto_id': ASSIGNED_TO_ID,
               'blocked_count': BLOCKED_COUNT,
               'case_ids': [CASE1_ID, CASE2_ID],
               'completed_on': COMPLETED_ON,
               'config': None,
               'config_ids': CONFIGS,
               'created_by': CREATED_BY_ID,
               'created_on': CREATED_ON,
               'custom_status1_count': 11,
               'custom_status2_count': 22,
               'custom_status3_count': 33,
               'custom_status4_count': 44,
               'custom_status5_count': 55,
               'custom_status6_count': 66,
               'custom_status7_count': 77,
               'description': DESCRIPTION,
               'failed_count': FAILED_COUNT,
               'id': ID,
               'include_all': False,
               'is_completed': True,
               'milestone_id': MILESTONE_ID,
               'name': NAME,
               'passed_count': PASSED_COUNT,
               'plan_id': PLAN_ID,
               'project_id': PROJECT_ID,
               'retest_count': RETEST_COUNT,
               'suite_id': SUITE_ID,
               'untested_count': UNTESTED_COUNT,
               'url': URL}
    return Run(client, deepcopy(content))


def test_assigned_to_get(new_run, run):
    run.client.api.user_by_id.return_value = {'id': ASSIGNED_TO_ID}

    assert new_run.assigned_to is None
    assert isinstance(run.assigned_to, User)
    assert run.assigned_to.id == ASSIGNED_TO_ID

    run.client.api.user_by_id.assert_called_with(ASSIGNED_TO_ID)


def test_assigned_to_set(client, new_run, run):
    user1 = User(client, {'id': ASSIGNED_TO_ID})
    user2 = User(client, {'id': ASSIGNED_TO_ID + 1})

    run.client.api.user_by_id.return_value = {'id': ASSIGNED_TO_ID}

    assert new_run.assigned_to is None
    new_run.assigned_to = user1
    assert new_run._content['assignedto_id'] == ASSIGNED_TO_ID

    assert run.assigned_to.id == ASSIGNED_TO_ID
    run.assigned_to = user2
    assert run._content['assignedto_id'] == ASSIGNED_TO_ID + 1

    run.client.api.user_by_id.assert_called_with(ASSIGNED_TO_ID)


def test_assigned_to_set_exc(client, new_run, run):

    assert new_run.assigned_to is None

    with pytest.raises(TypeError) as exc:
        new_run.assigned_to = 1234

    assert str(User) in str(exc)
    assert str(int) in str(exc)


def test_blocked_count(new_run, run):
    assert new_run.blocked_count is None
    assert run.blocked_count == BLOCKED_COUNT


def test_cases_get(new_run, run):
    run.client.api.case_by_id.side_effect = [CASE1, CASE2]

    assert list(new_run.cases) == list()

    cases = run.cases
    case1 = next(cases)
    assert isinstance(case1, Case)
    assert case1.id == CASE1_ID

    case2 = next(cases)
    assert isinstance(case2, Case)
    assert case2.id == CASE2_ID

    assert call(CASE1_ID) in run.client.api.case_by_id.call_args_list
    assert call(CASE2_ID) in run.client.api.case_by_id.call_args_list


def test_cases_set_list(new_run):
    assert list(new_run.cases) == list()
    new_run.cases = list()
    assert list(new_run.cases) == list()
    new_run.cases = [Case(None, CASE1), Case(None, CASE2)]
    assert new_run._content['case_ids'] == [CASE1_ID, CASE2_ID]


def test_cases_set_set(new_run):
    assert set(new_run.cases) == set()
    new_run.cases = set()
    assert set(new_run.cases) == set()
    new_run.cases = {Case(None, CASE1), Case(None, CASE2)}
    assert new_run._content['case_ids'].count(CASE1_ID) == 1
    assert new_run._content['case_ids'].count(CASE2_ID) == 1


def test_cases_set_tuple(new_run):
    assert tuple(new_run.cases) == tuple()
    new_run.cases = tuple()
    assert tuple(new_run.cases) == tuple()
    new_run.cases = (Case(None, CASE1), Case(None, CASE2))
    assert new_run._content['case_ids'] == [CASE1_ID, CASE2_ID]


def test_cases_set_generator(new_run):
    assert list(new_run.cases) == list()
    new_run.cases = (x for x in range(1) if x == 'a')  # an empty generator
    assert list(new_run.cases) == list()
    new_run.cases = (c for c in [Case(None, CASE1), Case(None, CASE2)])
    assert new_run._content['case_ids'] == [CASE1_ID, CASE2_ID]


def test_cases_set_list_exc(new_run):
    assert list(new_run.cases) == list()

    with pytest.raises(TypeError) as exc:
        new_run.cases = "1, 2"

    assert 'generator). Found {0}'.format(str) in str(exc)


def test_cases_set_type_exc(new_run):
    assert list(new_run.cases) == list()

    with pytest.raises(TypeError) as exc:
        new_run.cases = [Case(None, CASE1), '2']

    assert 'Found at least one {0}'.format(str) in str(exc)


def test_cases_set_type_exc_2(new_run):
    assert list(new_run.cases) == list()

    with pytest.raises(TypeError) as exc:
        new_run.cases = [CASE1_ID, CASE2]

    assert 'Found at least one {0}'.format(dict) in str(exc)


def test_completed_on(new_run, run):
    assert new_run.completed_on is None
    assert isinstance(run.completed_on, datetime.datetime)


def test_configs(new_run, run):
    assert list(new_run.configs) == list()

    run.client.api.project_by_id.return_value = {'id': PROJECT_ID}
    run.client.api.config_groups.return_value = [CG1]

    configs = run.configs

    conf1 = next(configs)
    assert isinstance(conf1, Config)
    assert conf1.id == CONF1_ID

    conf2 = next(configs)
    assert isinstance(conf2, Config)
    assert conf2.id == CONF2_ID

    run.client.api.project_by_id.assert_called_with(PROJECT_ID)
    run.client.api.config_groups.assert_called_with(PROJECT_ID)


def test_created_by(new_run, run):
    run.client.api.user_by_id.return_value = {'id': CREATED_BY_ID}

    assert new_run.created_by is None
    assert isinstance(run.created_by, User)
    assert run.created_by.id == CREATED_BY_ID

    run.client.api.user_by_id.assert_called_with(CREATED_BY_ID)


def test_created_on(new_run, run):
    assert new_run.created_on is None
    assert isinstance(run.created_on, datetime.datetime)


def test_description_get(new_run, run):
    assert new_run.description is None
    assert run.description == DESCRIPTION


def test_description_set(new_run, run):
    new_run.description = "New description 1"
    assert new_run.description == "New description 1"

    run.description = "New description 2"
    assert run.description == "New description 2"


def test_description_set_exc(new_run):
    with pytest.raises(TypeError) as exc:
        new_run.description = 1234

    assert str(str) in str(exc)
    assert str(int) in str(exc)


def test_failed_count(new_run, run):
    assert new_run.failed_count is None
    assert run.failed_count == FAILED_COUNT


def test_include_all(new_run, run):
    assert new_run.include_all is True
    assert run.include_all is False


def test_include_all_set(run):
    assert run.include_all is False

    run.include_all = True

    assert run.include_all is True


def test_include_all_set_exc(new_run):
    with pytest.raises(TypeError) as exc:
        new_run.include_all = 1234

    assert str(bool) in str(exc)
    assert str(int) in str(exc)


def test_is_completed(new_run, run):
    assert new_run.is_completed is False
    assert run.is_completed is True


def test_milestone_get(new_run, run):
    assert new_run.milestone is None

    run.client.api.milestone_by_id.return_value = {'id': MILESTONE_ID}

    assert isinstance(run.milestone, Milestone)
    assert run.milestone.id == MILESTONE_ID
    run.client.api.milestone_by_id.assert_called_with(MILESTONE_ID)


def test_milestone_set(new_run):
    assert new_run.milestone is None

    new_run.milestone = Milestone(None, {'id': MILESTONE_ID})

    assert new_run._content['milestone_id'] == MILESTONE_ID


def test_milestone_set_exc(new_run):
    assert new_run.milestone is None

    with pytest.raises(TypeError) as exc:
        new_run.milestone = 1234

    assert str(Milestone) in str(exc)
    assert str(int) in str(exc)


def test_name_get(new_run, run):
    assert new_run.name is None
    assert run.name is NAME


def test_name_set(new_run, run):
    new_run.name = NAME
    assert new_run.name == NAME

    run.name = NAME + " addition"
    assert run.name == NAME + " addition"


def test_name_set_exc(new_run):
    with pytest.raises(TypeError) as exc:
        new_run.name = 1234

    assert str(str) in str(exc)
    assert str(int) in str(exc)


def test_passed_count(new_run, run):
    assert new_run.passed_count is None
    assert run.passed_count == PASSED_COUNT


def test_plan(new_run, run):
    assert new_run.plan is None

    run.client.api.plan_by_id.return_value = {'id': PLAN_ID}

    assert isinstance(run.plan, Plan)
    assert run.plan.id == PLAN_ID
    run.client.api.plan_by_id.assert_called_with(PLAN_ID)


def test_project_get(new_run, run):
    run.client.api.project_by_id.return_value = {'id': PROJECT_ID}

    assert new_run.project is None
    assert isinstance(run.project, Project)

    assert run.project.id == PROJECT_ID
    run.client.api.project_by_id.assert_called_with(PROJECT_ID)


def test_project_set(new_run):
    project = Project(None, {'id': PROJECT_ID})
    new_run.project = project

    assert new_run._content['project_id'] == PROJECT_ID


def test_project_set_exc(new_run):
    with pytest.raises(TypeError) as exc:
        new_run.project = 1234

    assert str(Project) in str(exc)
    assert str(int) in str(exc)


def test_retest_count(new_run, run):
    assert new_run.retest_count is None
    assert run.retest_count == RETEST_COUNT


def test_suite_get(new_run, run):
    run.client.api.suite_by_id.return_value = {'id': SUITE_ID}

    assert new_run.suite is None
    assert isinstance(run.suite, Suite)

    assert run.suite.id == SUITE_ID
    run.client.api.suite_by_id.assert_called_with(SUITE_ID)


def test_suite_set(new_run):
    suite = Suite(None, {'id': SUITE_ID})
    new_run.suite = suite

    assert new_run._content['suite_id'] == SUITE_ID


def test_suite_set_exc(new_run):
    with pytest.raises(TypeError) as exc:
        new_run.suite = 1234

    assert str(Suite) in str(exc)
    assert str(int) in str(exc)


def test_untested_count(new_run, run):
    assert new_run.untested_count is None
    assert run.untested_count == UNTESTED_COUNT


def test_url(new_run, run):
    assert new_run.url is None
    assert isinstance(run.url, str)
    assert run.url == URL



from copy import deepcopy
import logging
import time

import requests
from retry import retry
from requests.status_codes import codes
from requests.exceptions import ChunkedEncodingError, ConnectionError

from .const import BASE_API_PATH, TIMEOUT
from .exceptions import (BadRequest, Conflict, Forbidden, NotFound, RateLimited,
                         Redirect, ServerError, TooLarge, UnknownStatusCode)

log = logging.getLogger(__package__)


class Session(object):
    """  """
    RATE_LIMIT_STATUS = 429
    RETRY_EXCEPTIONS = (ChunkedEncodingError, ConnectionError, RateLimited, ServerError)
    STATUS_EXCEPTIONS = {codes['bad_gateway']: ServerError,
                         codes['bad_request']: BadRequest,
                         codes['conflict']: Conflict,
                         codes['found']: Redirect,
                         codes['forbidden']: Forbidden,
                         codes['gateway_timeout']: ServerError,
                         codes['internal_server_error']: ServerError,
                         codes['not_found']: NotFound,
                         codes['request_entity_too_large']: TooLarge,
                         codes['service_unavailable']: ServerError,
                         codes['unauthorized']: Forbidden}
    SUCCESS_STATUSES = {codes['created'], codes['ok']}

    def __init__(self, auth, url):
        """ Prepare the connection to the TestRail API

        :param auth: Tuple of username and api_key/password
        :param url: Base url for testrail (e.g. https://<your company>.testrail.net)

        """
        self._auth = auth
        self._url = url

        self._http = requests.Session()
        self._http.headers['Content-Type'] = 'application/json'

    @staticmethod
    def _log_request(**kwargs):
        method = kwargs.get('method', None)
        url = kwargs.get('url', None)
        json = kwargs.get('json', None)
        params = kwargs.get('params', None)
        log.debug('Fetching: {} {}'.format(method, url))
        log.debug('JSON    : {}'.format(json))
        log.debug('Params  : {}'.format(params))

    def _make_request(self, *args, **kwargs):
        kwargs['timeout'] = TIMEOUT
        kwargs['auth'] = self._auth
        try:
            response = self._http.request(*args, **kwargs)
            log.debug('Response: {} ({} bytes)'.format(
                response.status_code, response.headers.get('content-length')))
        except Exception as exc:
            return None, exc
        else:
            return response, None

    @retry(RETRY_EXCEPTIONS, tries=3, delay=1, backoff=2)
    def _request_with_retries(self, *args, **kwargs):
        """  """
        self._log_request(**kwargs)
        response, _ = self._make_request(*args, **kwargs)

        if response.status_code in self.STATUS_EXCEPTIONS:
            raise self.STATUS_EXCEPTIONS[response.status_code](response)
        elif response.status_code == codes['no_content']:
            return
        elif response.status_code == self.RATE_LIMIT_STATUS:
            retry_after = int(response.headers['Retry-After'])
            log_msg = 'API rate limit reached. Retrying after {0} seconds'
            log.warning(log_msg.format(retry_after))
            time.sleep(retry_after)
            raise RateLimited(response)
        elif response.status_code not in self.SUCCESS_STATUSES:
            raise UnknownStatusCode(response)

        if response.headers.get('content-length') == '0':
            return ''
        else:
            return response.json()

    def close(self):
        """ Close the session """
        self._http.close()

    def request(self, method, path, json=None, params=None):
        """Return the json content from the resource at ``path``.

        :param method: The request verb. E.g., get, post, put.
        :param path: The path of the request. This path will be combined with
            the base TestRail api URL.
        :param json: Object to be serialized to JSON in the body of the
            request.
        :param params: The query parameters to send with the request.

        """
        params = deepcopy(params) or dict()
        url = '/'.join(part.strip('/') for part in [self._url, BASE_API_PATH, path])
        return self._request_with_retries(method=method, json=json, params=params, url=url)



from .model_base import ModelBase


class CaseType(ModelBase):
    """ Object model for TestRail Case Types

    To get all case types

    .. code-block:: python

        case_types = list(traw_client.case_types())

    """
    @property
    def is_default(self):
        """ True if the case type is the default """
        return self._content.get('is_default', None)

    @property
    def name(self):
        """ The full name of the case type """
        return self._content.get('name', None)



"""  TRAW: TestRail API Wrapper

TRAW is an API wrapper for Gurrock's TestRail test management suite

The intended way to begin is to instantiate the TRAW Client:

.. code-block:: python

    import traw
    testrail = traw.Client(username='username',
                           user_api_key='api_key',
                           url='url')

See the Client help documentation (`help(traw.Client)`) for more information
"""
import logging

from pbr.version import VersionInfo

from .client import Client  # NOQA


__version__ = VersionInfo('traw').semantic_version().release_string()
__all__ = ('__version__', 'Client')

logging.getLogger(__package__).addHandler(logging.NullHandler())



from .model_base import ModelBase


class Status(ModelBase):
    """ Object model for TestRail Statuses

    To get all statuses

    .. code-block:: python

        statuses = list(traw_client.statuses())

    """
    @property
    def color_bright(self):
        """ The brightest shade of the status """
        return self._content.get('color_bright')

    @property
    def color_dark(self):
        """ The darkest shade of the status """
        return self._content.get('color_dark')

    @property
    def color_medium(self):
        """ The middle shade of the status """
        return self._content.get('color_medium')

    @property
    def is_final(self):
        """ True if this is a final status (e.g. not an intermediary status) """
        return self._content.get('is_final')

    @property
    def is_system(self):
        """ True if this is a system status (e.g. not a custom status) """
        return self._content.get('is_system')

    @property
    def is_untested(self):
        """ True if the status indicates an untested state """
        return self._content.get('is_untested')

    @property
    def label(self):
        """ The displayed label of the status """
        return self._content.get('label', None)

    @property
    def name(self):
        """ The name of the status """
        return self._content.get('name', None)
